(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 *   Revised for SML 110 - 5/98
 *
 * General code to manage debugging options
 *)

structure U =
    struct
	datatype 'a desc = N | FREE | BOUND of 'a
	fun rappend nil y = y | rappend (x::y) z = rappend y (x::z)

	fun foldstate f =
	    let
		fun fs (nil, b) = (nil, b)
		  | fs (e::l,b) =
		    let val (e',b1) = f(e,b)
			val (l',b2) = fs (l,b1)
		    in (e'::l', b2)
		    end
	    in fs
	    end

	fun member (x,nil) = false
	  | member (x,a::r) = x = a orelse member(x,r)
    end

structure Debug =
    struct
	(*
	 * debug option:
	 *    key: how to activate/deactivate/test option
	 *    purpose: what activating the option will do
	 *    value: is the option now on?
	 *)
	datatype debug = 
	    DB of {key: string,
		   purpose: string,
		   value: bool ref}

	(* debugging options being maintained *)
	val debugs = ref (nil: debug list)
		   
	(* reset debugging options *)
	fun init () = (debugs := nil)

	(* make a new debugging key *)
	fun register (key, purpose, default) =
	    let
		val hids = 
		    (* ListUtil.find *)
		    List.filter 
		    (fn DB{key=k,...} => key=k) (!debugs)
	    in
		case hids of 
		    nil => 
			debugs := DB{key=key, 
				     purpose=purpose, 
				     value=ref default} :: (!debugs)
		  | l => 
			(print "WARNING: Key ";
			 print key;
			 print " overlaps with options with these purposes:\n";
			 app (fn DB{purpose,...} => print purpose) l;
			 debugs := DB{key=key,
				      purpose=purpose,
				      value = ref default} :: (!debugs))
	    end (* let *)

	(* debug: turn debugging of ARG on *)
	fun debug "help" =
	    (app (fn DB{key,purpose,...} => (print key;
					     print ": ";
					     print purpose))
	     (!debugs))
	     
	  | debug "status" =
	    (app (fn DB{key,value,...} => (print key;
					   print ": ";
					   print 
					   ((* makestring *)
					    Bool.toString 
					    (!value))))
	     (!debugs))
	     
	  | debug str =
	    let
		val item = 
		    (* ListUtil.findOne *)
		    List.find 
		    (fn DB{key=k,...} => str=k) (!debugs)
	    in
		case item of
		    NONE => 
			(print "WARNING: no such option ";
			 print str;
			 print "\n")
		  | SOME (DB{value,...}) => 
			value := true
	    end (* let *)

	(* nodebug: turn it off *)
	fun nodebug str =
	    let
		val item = 
		    (* ListUtil.findOne *)
		    List.find
		    (fn DB{key=k,...} => str=k) (!debugs)
	    in
		case item of
		    NONE => 
			(print "WARNING: no such option ";
			 print str;
			 print "\n")
		  | SOME (DB{value,...}) => 
			value := false
	    end (* let *)

	(* debugp: get whether thing is on *)
	fun debugp str =
	    let
		val item = 
		    (* ListUtil.findOne *)
		    List.find
		    (fn DB{key=k,...} => str=k) (!debugs)
	    in
		case item of
		    NONE => 
			(print "WARNING: no such option ";
			 print str;
			 print "\n";
			 false)
		  | SOME (DB{value,...}) => 
			(!value)
	    end (* let *)
	    
	fun getline () = 
	    (* input_line std_in *)
	    TextIO.inputLine TextIO.stdIn
	    
    end (* struct *)
 
