(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 *
 * Abstraction to time expressions and function applications.
 *)

signature CLOCK =
    sig
	(* time F() -- with name STRING --- and print time elapsed *)
	val clocked : (unit -> 'a) -> string -> 'a

	(* build function that applies F -- with name STRING -- 
	 * and prints time elapsed *)
	val clockapp : ('b -> 'a) -> string -> 'b -> 'a

	(* allow something to be interrupted *)
	val interruptible : (unit -> '1a) * '1a -> '1a
    end

structure Clock : CLOCK =
    struct
	fun myview t = 
	    (LargeInt.toString (Time.toMilliseconds t)) ^ " msec"

	(* time F() -- with name STRING --- and print time elapsed *)
	fun clocked f (announce:string) =
	    let
		val startTime = Time.now()
		val t = Timer.startCPUTimer()

		val result = f()

		val elapsedTime = Time.- (Time.now(), startTime)
		val elapsedCPUTime = Timer.checkCPUTimer(t)
	    in
		print ("[ " ^
		       announce ^
		       "  \ntime elapsed: " ^
		       (myview (#usr elapsedCPUTime)) ^ " user/ " ^
		       (myview (#sys elapsedCPUTime)) ^ " sys/ " ^
		       (myview (#gc elapsedCPUTime)) ^ " gc/ " ^
		       (myview elapsedTime) ^ " wall ]\n");
		result
	    end

	(* build function that applies F -- with name STRING -- 
	 * and prints time elapsed *)
	fun clockapp f (announce:string) b =
	    let
		val startTime = Time.now()
		val t = Timer.startCPUTimer()

		val result = f b

		val elapsedTime = Time.- (Time.now(), startTime)
		val elapsedCPUTime = Timer.checkCPUTimer(t)
	    in
		print ("[ " ^
		       announce ^
		       "  \ntime elapsed: " ^
		       (myview (#usr elapsedCPUTime)) ^ " user/ " ^
		       (myview (#sys elapsedCPUTime)) ^ " sys/ " ^
		       (myview (#gc elapsedCPUTime)) ^ " gc/ " ^
		       (myview elapsedTime) ^ " wall ]\n");
		result
	    end

	fun interruptible (f, r) =
	    let 
		val oldHandler = Signals.inqHandler(Signals.sigINT)
		val result = ref r
	    in
		SMLofNJ.Cont.callcc
		(fn abort =>
		       let fun handler _ = (print "Interrupted\n"; abort)
		       in  Signals.setHandler(Signals.sigINT, Signals.HANDLER handler);
			   result := f();
			   ignore (Signals.setHandler(Signals.sigINT, oldHandler))
		       end);
		!result
	    end
    end

	    
