
functor FieldStrF (structure L : LEX) : FIELD_STR =
    struct
	datatype 'a report =
	    GOT of 'a
	  | EMPTY
	  | ERROR of string 

	datatype 'a fld =
	    F of {name: string,
		  read: string list -> 'a option * string list}

	datatype 'a node =
	    Tree of {key: string,
		     data: 'a,
		     kids: 'a node list}

	fun merge nil = ""
	  | merge (a :: more) = 
	    (case merge more of
		 "" => a
	       | x => a ^ " " ^ x)

	fun close "(" = SOME ")" 
	  | close "{" = SOME "}"
	  | close "[" = SOME "]" 
	  | close _ = NONE

	fun glom (_, _, sofar, nil) = NONE
	  | glom (closer, more, sofar, thing :: rest) =
	    if closer = thing then
		finish (more, sofar, rest, thing)
	    else case close thing of
		NONE => glom (closer, more, thing :: sofar, rest) 
	      | SOME newclose => glom (newclose, closer :: more,
				       thing :: sofar,
				       rest)

	and finish (nil, sofar, rest, _) = SOME (rev sofar, rest)
	  | finish (top :: more, sofar, rest, thing) =
	    glom(top, more, thing :: sofar, rest)

	fun recread (name, f, msg)  (all as (str :: "=" :: openp :: stuff)) =
	    if name = str then
		(case close openp of
		     SOME closer =>
			 (case glom (closer, nil, nil, stuff) of
			      SOME (sofar, more) => (f sofar, more)
			    | NONE => 
				  ((ERROR(msg ^ " parentheses don't match\n"), all)))
		   | NONE => 
			 ((ERROR(msg ^ " data not enclosed in parens\n"), all)))
	    else
		((ERROR("name error: expected " ^ name ^ "; got " ^ str^ "\n"),
		  all))
	  | recread _ nil = ((EMPTY, nil))
	  | recread (name, _, msg) all =
	    ((ERROR(msg ^ " doesn't begin \"" ^ name ^ " = " ^ 
		    (Char.toString (chr(123))) ^ " ...\" \n"),  all))
	    
	local
	    fun transform s (SOME v, nil) = (GOT v)
	      | transform s (SOME v, _) = ERROR(s ^ " not all consumed\n")
	      | transform s (NONE,_) = ERROR (s ^ "couldn't read data\n")
	in
	    fun read (F{name, read, ...}) = 
		let 
		    val msg = "field " ^ name
		in
		    recread (name, (transform msg) o read, msg)
		end
	end

	fun readall fields nil = EMPTY
	  | readall fields strings =
	    let
		exception error of string
		fun next (f, (s, a)) =
		    case read f s of
			(GOT guy, rest) => (rest, guy :: a)
		      | (ERROR s, _) => raise error s
		      | (EMPTY, _) => (raise error "premature end of file\n")
	    in
		(case (foldl next (strings, nil) fields) of
		     (nil, guys) =>
			 GOT (rev guys)
		   | _ =>
			 ERROR("leftover material after reading\n"))
		handle error s =>  ERROR s
	    end

	fun readstruct (name, fields) = recread(name, readall fields, "struct " ^ name)

	fun segall (rf, fields) nil = EMPTY
	  | segall (rf, fields) strings =
	    let
		exception error of string

		fun next (F{name,...}, (s, a)) =
		    case recread (name, rf, "field " ^ name) s of
			(GOT guy, rest) => (rest, guy :: a)
		      | (ERROR s, _) => raise (error s)
		      | (EMPTY, _) => (raise error "premature end of file\n")
	    in
		(case (foldl next (strings, nil) fields) of
		     (nil, guys) =>
			 GOT (rev guys)
		   | _ =>
			 ERROR ("leftover material after reading\n"))
		handle error s => ERROR s
	    end

	fun rawSegment (name, fields) = recread(name, segall (GOT, fields), "struct " ^ name)
	fun segment (name, fields) = recread(name, segall (GOT o merge, fields), "struct " ^ name)

	fun unsegment (name, fields) data =
	    let
		fun dowrite (F{name,...} :: moref, value :: morev, s) =
		    "\n   " ^ name ^ " = " ^ (Char.toString (chr(123))) ^ value ^ 
		    (Char.toString (chr(125))) ^ 
		    (dowrite(moref, morev, s))
		  | dowrite (_, _, s) = s
	    in
		name ^ " = " ^ (Char.toString (chr(123))) ^ 
		dowrite(fields, data, "\n" ^ (Char.toString (chr(125))))
	    end

	fun test (field as F{name,read,...}) s = 
	    (case (read (L.segment s)) of
		 (SOME guy, nil) => true
	       | (SOME _, more) => 
		     (print ("field " ^ name ^ " data not all consumed\n");
		      false)
	       | (NONE, _) =>
		     (print ("field " ^ name ^ " couldn't read data\n");
		      false))

	fun readBool ("t" :: more) = (SOME true, more)
	  | readBool ("T" :: more) = (SOME true, more)
	  | readBool ("yes" :: more) = (SOME true, more)
	  | readBool ("Yes" :: more) = (SOME true, more)
	  | readBool ("YES" :: more) = (SOME true, more)
	  | readBool ("true" :: more) = (SOME true, more)
	  | readBool ("True" :: more) = (SOME true, more)
	  | readBool ("TRUE" :: more) = (SOME true, more)

	  | readBool ("f" :: more) = (SOME false, more)
	  | readBool ("F" :: more) = (SOME false, more)
	  | readBool ("no" :: more) = (SOME false, more)
	  | readBool ("No" :: more) = (SOME false, more)
	  | readBool ("NO" :: more) = (SOME false, more)
	  | readBool ("false" :: more) = (SOME false, more)
	  | readBool ("False" :: more) = (SOME false, more)
	  | readBool ("FALSE" :: more) = (SOME false, more)
	    
	  | readBool s = (NONE, s)

	fun readInt (s :: more) = (Int.fromString s, more)
	  | readInt x = (NONE, x)

	fun readStrings s = (SOME s, nil)

	fun wrapReader (f, r) s =
	    let
		val (a,s') = r s
	    in
		case a of 
		    NONE => (NONE, s')
		  | SOME a' => (f a', s')
	    end

	fun readListOf f s =
	    let
		exception no

		fun rf (nil, sofar) = (SOME (rev sofar), nil)
		  | rf ("," :: more, sofar) = rf (more, sofar)
		  | rf (";" :: more, sofar) = rf (more, sofar)
		  | rf (s, sofar) =
		    (case f s of
			 (SOME n, s') => rf (s', n :: sofar)
		       | _ => raise no)
	    in
		rf (s, nil) 
		handle no => (NONE, s)
	    end
    end

structure FieldStr = FieldStrF(structure L = Lex1);
