(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 *
 * INTFIFO
 *   Funny priority queue needed for tree construction
 *   implementation.
 * Each entry is associated with an integer (later
 *   used to represent the depth in the tree at which
 *   a queued merge is to be done).
 * Entries have FIFO discipline by code, but lower
 *   coded entries are always picked before higher
 *   coded entries.
 *)

signature INTFIFO =
  sig
    type 'a fifo
    datatype emptylevel = Empty | EmptyUpTo of int

    exception Dequeue

    (* return empty queue *)
    val empty : unit -> 'a fifo

    (* test a queue for emptiness *)
    val isEmpty : 'a fifo -> bool

    (* return the lowest code with a nonempty subqueue *)
    val howEmpty : 'a fifo -> emptylevel

    (* add element to queue with priority int *)
    val enqueue : 'a fifo * 'a * int -> 'a fifo
	
    (* remove element from queue *)
    val dequeue : 'a fifo -> 'a fifo * 'a

    (* get -- but don't remove -- element from queue *)
    val head : 'a fifo -> 'a

    (* return number of elements of queue *)
    val length : 'a fifo -> int

    (* flatten queue into a list *)
    val contents : 'a fifo -> 'a list
	
    (* apply function to all elements of queue, forward... *)
    val app : ('a -> 'b) -> 'a fifo -> unit
	
    (* ...and backward *)
    val revapp : ('a -> 'b) -> 'a fifo -> unit

    (* transform the queue *)
    val map : ('a -> 'b) -> 'a fifo -> 'b fifo

    (* sum over the queue, forward... *)
    val fold : ('a * 'b -> 'b) -> 'b -> 'a fifo -> 'b

    (* and backward ... *)
    val revfold : ('a * 'b -> 'b) -> 'b -> 'a fifo -> 'b

  end (* FIFO *)
