(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 *
 * INTFIFO implementation.
 *    Uses my modified int maps with their fast
 *    function for getting the next key.
 *)

functor IntMapIntFifo (structure IntMap : MYINTMAP;
		       structure Fifo : FIFO) : INTFIFO =
    struct
	datatype emptylevel = Empty | EmptyUpTo of int

	datatype 'a fifo = 
	    F of 'a Fifo.fifo IntMap.intmap * emptylevel

	exception Dequeue

	fun empty () = F(IntMap.empty(), Empty)

	fun isEmpty (F(_,Empty)) = true
	  | isEmpty _ = false

	fun howEmpty (F(_,e)) = e

	fun q(s,l) = 
	    case IntMap.peek(s,l) of 
		NONE => Fifo.empty | SOME q => q

	fun enqueue (F(s,e), elt, level) =
	    let
		val news = 
		    IntMap.insert(s, level, Fifo.enqueue(q(s,level), elt))
		val newe = 
		    case e of Empty => EmptyUpTo level
		      | EmptyUpTo l => EmptyUpTo(
						 (* min *) Int.min
						 (l,level))
	    in
		F(news, newe)
	    end

	fun dequeue (F(s,Empty)) = raise Dequeue
	  | dequeue (F(s,EmptyUpTo d)) =
	    let val (newq, r) = Fifo.dequeue (q(s,d))
	    in if Fifo.isEmpty newq then
		let val (news, _) = IntMap.remove(s, d) 
		    handle IntMap.NotFound => 
			raise (Fail ("dequeue (int-fifo): " ^
				     (* (makestring d) ^ *)
				     (Int.toString d) ^
				     " not found"))
		    val newe = 
			case IntMap.next(news,d) of
			    NONE => Empty
			  | SOME(d,_) => EmptyUpTo d
		in (F(news, newe),r) end
	       else (F(IntMap.insert(s, d, newq), EmptyUpTo d), r) 
	    end

	fun head (F(s,Empty)) = raise Dequeue
	  | head (F(s,EmptyUpTo d)) = Fifo.head(q(s,d))

	fun length (F(s,_)) =
	    IntMap.fold (fn(v,q,l) => Fifo.length q + l) s 0

	fun contents (F(s,_)) =
	    IntMap.fold (fn(v,q,l) => (Fifo.contents q) @ l) s nil

	fun app f (F(s,_)) =
	    IntMap.app (fn (v,q) => Fifo.app (fn x => (f x; ())) q) s

	fun revapp f (F(s,_)) =
	    IntMap.revapp (fn (v,q) => Fifo.foldl (fn (u,_) => (f u; ())) () q) s

	fun map f (F(s,e)) =
	    F(IntMap.transform (Fifo.map f) s, e)

	fun fold f v (F(s,_)) =
	    IntMap.fold (fn (n,q,r) => Fifo.foldr f r q) s v

	fun revfold f v (F(s,_)) =
	    IntMap.revfold (fn (n,q,r) => Fifo.foldl f r q) s v

    end (* struct *)
