
signature DERIVE =
    sig
	structure Itag : ITAG;
	structure Word : WORD;
	structure Constraint : CONSTRAINT;
	sharing Constraint.Call = Itag.Call = Word.Call;
	sharing Word.Formula = Constraint.Formula = Itag.Formula;

	type deriveState
	type generand
	datatype generand_part =
	    NAME of string
	  | FMLA of Constraint.Formula.formula
	  | CALL of Constraint.Call.call
	  | FEATURES of Itag.Tag.Features.features
	  | GOALS of Constraint.Formula.formula list
	  | TERM of Constraint.Call.Term.term
	  | KEY of int

	val generandFields : (Word.Formula.Patheq.Modspec.collection ->
			      string -> 
			      generand_part FieldStr.fld list)

	val gfieldInits : string -> string list

	val readGenerand :  (Word.Formula.Patheq.Modspec.collection ->
			     string list -> generand option * string list)
	val assembleGenerand : string -> generand_part list -> generand option

	val start : (Itag.Tag.Features.features *
		     Itag.Call.call *
		     Itag.Call.call *
		     Word.Formula.formula list *
		     (Word.Formula.Patheq.pathpt * (Word.Formula.Term.init Word.Formula.Term.IntMap.intmap  * 
						   Word.Formula.formula -> 
						   unit Word.Formula.proof list))) -> deriveState

	val printState : deriveState -> unit
	val prettyprintState : ((string * Itag.Tag.Features.features * Itag.Tag.Features.binding ->
				 string option) ->
				deriveState -> unit)
	val outputResult : ((string * Itag.Tag.Features.features * Itag.Tag.Features.binding ->
			     string option) * (string -> unit) ->
			    deriveState list -> unit)
	val remainingGoals : deriveState list -> Word.Formula.formula list

	val stateOutFields : string FieldStr.fld list
	val writeState : ((string * Itag.Tag.Features.features * Itag.Tag.Features.binding ->
				 string option) ->
			  deriveState -> (string * string list) FieldStr.node * string list)

	val extensions :
	    (Word.lexicon * Itag.taglex *
	     (Word.Formula.Patheq.pathpt * (Word.Formula.Term.init Word.Formula.Term.IntMap.intmap  * 
				       Word.Formula.formula -> 
				       unit Word.Formula.proof list)) *
	     Word.Formula.formula * 
	     (Word.Formula.Patheq.pathpt * (Word.Formula.Term.init Word.Formula.Term.IntMap.intmap  *
				       Word.Formula.formula ->
				       unit Word.Formula.proof list))) -> 
	     deriveState -> deriveState list

	val completeness : deriveState -> int option

	val expandAbbrev : (Word.Formula.database *
			    Word.Formula.formula *
			    Word.Formula.formula *
			    Word.Formula.formula) ->
	    Word.Formula.formula list

	val generateList : (Word.lexicon * Itag.taglex * Word.Formula.database -> 
			    generand -> deriveState list)

    end
