
functor Derive (structure Itag : ITAG;
		structure Word : WORD;
		structure Constraint : CONSTRAINT;

		sharing Constraint.Call = Itag.Call = Word.Call;
		sharing Word.Formula = Itag.Formula = Constraint.Formula;
		sharing Word.Tag = Itag.Tag) : DERIVE =
    struct
	structure Itag = Itag;
	structure Word = Word;
	structure Constraint = Constraint;
	    
	(* more or less from paulson p 98;
	  cmp (A,B) when A should occur strictly before B *)

	fun quicksort cmp =
	    let
		fun quick ([],more) = more
		  | quick ([x],more) = x :: more
		  | quick (a::bs,more) =
		    let
			fun part (left, right, nil) =
			    quick (rev left, a :: (quick (rev right, more)))
			  | part (left, right, x::xs) =
			    if cmp (x,a) then 
				part(x::left, right, xs)
			    else
				part(left, x::right, xs)
		    in
			part(nil,nil,bs)
		    end
	    in
		fn x => quick(x,nil)
	    end

	datatype score_info =
	    SI of {cgoals: Constraint.Formula.formula list,
		   issubst: bool,
		   basic: bool,
		   sizes: int list,
		   work: int list,
		   numfree: int,
		   incompleteness: int}

	datatype deriveState = 
	    DS of {constraint: Constraint.problem,

		   specinfo: Constraint.Call.info,
		   spectinst: Itag.Tag.tinst,

		   geninfo: Constraint.Call.info,
		   gentinst: Itag.Tag.tinst,

		   semantics: Constraint.Formula.formula list,

		   score: score_info}

	fun mkScore (cgoals, issubst, basic, 
		     sizes, work, numfree, incompleteness) =
	    SI{cgoals = cgoals,
	       issubst = issubst,
	       basic = basic,
	       sizes = quicksort op> sizes,
	       work = quicksort op> work,
	       numfree = numfree,
	       incompleteness = incompleteness}

	fun printScore (SI{cgoals, issubst, basic, 
			   sizes, work, numfree, incompleteness}) =
	    let
		val _ =
		    if Debug.debugp "dosubst" then
			if issubst then
			    print "S: "
			else
			    print "X: "
		    else
			()
	    in
		(print (Int.toString (length cgoals));
		 print " ";
		 print (Int.toString (numfree));
		 print ";";
		 app (fn x => (print (Int.toString x); print " ")) sizes;
		 print "/ ";
		 app (fn x => (print (Int.toString x); print " ")) work;
		 print ";";
		 print (Int.toString incompleteness);
		 print " ";
		 print (if basic then "1" else "0"))
	    end

	fun cmpScore (DS{score=SI{cgoals=cg1, basic=b1, sizes=s1, work = w1,
				  numfree=n1, incompleteness=in1, ...},...},
		      DS{score=SI{cgoals=cg2, basic=b2, sizes=s2, work = w2,
				  numfree=n2, incompleteness=in2, ...},...}) =
	    let
		fun ones (nil) = true
		  | ones (1::m) = ones m
		  | ones _ = false

		fun c(l,nil) = if ones l then EQUAL else GREATER
		  | c(nil,l) = if ones l then EQUAL else LESS
		  | c(a::m, b::n) = 
		    if a < b then LESS
		    else if a = b then c(m,n) else GREATER

		val l1 = length cg1
		val l2 = length cg2
	    in
		l1 < l2 orelse
		(l1 = l2 andalso 
		 (n1 < n2 orelse
		  (n1 = n2 andalso 
		   (case c(s1,s2) of
			LESS => true
		      | GREATER => false
		      | EQUAL => 
			    (case c(w1,w2) of
				 LESS => true
			       | GREATER => false
			       | EQUAL => 
				     in1 < in2 orelse
				     (in1 = in2 andalso b1 andalso not b2))))))
	    end

	fun cmpIter (d1 as DS{score=SI{issubst,...},...}, d2) =
	    if Debug.debugp "dosubst" andalso issubst then
		    true
	    else
		cmpScore(d1,d2)
			
	fun completeness (DS{score=SI{cgoals,sizes,incompleteness,numfree,...},...}) =
	    if (incompleteness = 0 andalso
		numfree = 0 andalso
		not (List.exists (fn x => x > 1) sizes)) then 
		SOME (length cgoals)
	    else
		NONE

	fun finalize((privateworld,_),
		     ds as DS{constraint, specinfo, spectinst, geninfo, 
			      gentinst, semantics, score}) =
	    let
		fun inst f =
		    let
			val fvs = Constraint.Formula.freeVars f
			val testtr = Constraint.Call.getKeyInit (specinfo, fvs, privateworld)
		    in
			Constraint.Formula.subst (testtr, privateworld) f
		    end
	    in
		(case (Itag.Tag.finalize(spectinst), Itag.Tag.finalize(gentinst)) of
		     (SOME sf, SOME gf) =>
			 (DS{constraint = constraint,
			     specinfo = specinfo,
			     spectinst = sf,
			     geninfo = geninfo,
			     gentinst = gf,
			     semantics = map inst semantics,
			     score = score})
		   | _ => ds)
	    end

	fun start (features, speccall, gencall, cgoals, public) =
	    DS{constraint = Constraint.newProblem (public),

	       specinfo = Constraint.Call.mkInfo (),
	       spectinst = Itag.Tag.start (features, speccall),

	       geninfo = Constraint.Call.mkInfo (),
	       gentinst = Itag.Tag.start (features, gencall),

	       semantics = nil,
	       score = mkScore(cgoals, false, false, nil, nil,
			       99 (* make stopping bad *), 1)}
	       
	fun printState(d as DS{spectinst, constraint, specinfo, geninfo, score,...}) =
	    (Itag.Tag.printTinstLeaves spectinst;
	     print "score: ";
	     printScore score;
	     case completeness d of
		 NONE => 
		     print "\n"
	       | SOME 0 => 
		     print "(complete)\n"
	       | SOME x =>
		     print "(good sentence)\n")

	fun prettyprintState lookup (d as DS{spectinst, ...}) =
	    (Itag.Tag.prettyprintTinstLeaves lookup spectinst;
	     case completeness d of
		 NONE => 
		     print "\n"
	       | SOME 0 => 
		     print "(complete)\n"
	       | SOME x =>
		     print "(good sentence)\n")

	fun outputResult (lookup,_) nil = ()
	  | outputResult (lookup,output) ([d as DS{spectinst, semantics,...}]) =
	    (output (Itag.Tag.writeprettyTinstLeaves (lookup) spectinst);
	     output "\n";
	     app (fn x => (output "    ";
			   Constraint.Formula.printFormula output x;
			   output "\n")) semantics)
	  | outputResult p (_ :: r) = outputResult p r

	fun remainingGoals nil = nil
	  | remainingGoals ([DS{score=SI{cgoals,...},...}]) = cgoals
	  | remainingGoals (_ :: r) = remainingGoals r


	val stateOutFields = 
	    [FieldStr.F{name = "morphed leaves", read = fn s => (NONE,s)},
	     FieldStr.F{name = "raw leaves", read = fn s => (NONE, s)}] :
	    string FieldStr.fld list

	fun writeState lookup (DS{spectinst,specinfo,...}) =
	    (Itag.writerecEntry(spectinst, specinfo),
	     [Itag.Tag.writeprettyTinstLeaves lookup spectinst,
	      Itag.Tag.writeTinstLeaves spectinst] : string list)

	fun wout (s:string) =
	    if Debug.debugp "derive" then
		print s
	    else
		()

	fun notEntailed ((publicworld, publicset), 
			 shared, info, semantics) goal =
	    let
		val testf = 
		    Constraint.Formula.compose goal
		    (Constraint.Formula.substImplic
		     (semantics, Constraint.Formula.emptyCxt, 
		      shared, Constraint.Formula.emptyCxt))

		val fvs = Constraint.Formula.freeVars testf
		val testtr = Constraint.Call.getKeyInit (info, fvs, publicworld)
		val set = publicset(testtr,testf)
	    in
		case set of 
		    nil => true
		  | _ => false
	    end

	(*
	 * extensions: (lex) -> 
	 *    (tree * genTree * constraints) -> 
	 *    (tree * genTree * constraints) list
	 *
	 * Compute all the places you can go next from TREE, given
	 * lexicon LEX.
	 *
	 * For each place to act in TREE, find all the words
	 * that fit; then glom into a list all the TRUE instantiations
	 * of those words in triples (instantiation, type, opportunity).
	 * This list is stored in possWITOs.
	 *
	 * Then, find the different calls of trees that the different
	 * words invoke, and accumulate all the contextually appropriate
	 * instantiations of those trees.  This list is stored in 
	 * possTrees.
	 *
	 * Finally, unify the surviving trees with the surviving words
	 * and TAG them into the tree at the specified location.
	 *
	 * genTree gives the tree that you get by ignoring the 
	 * semantics, and constraints is the constraint problem
	 * associated with reference resolution.  These both
	 * have to be updated on the addition of new content.
	 *)

	fun assemble (public, sharedscheme, checkedtrees, lastsem, 
		      lastproblem, lastcgoals,
		      geninfo, lastspectree, lastgentree) 
	    (Word.M{name, basic, match, call, parms, init,
		    semantics, presup, trees, specinfo, 
		    oppo=Itag.Tag.OP{address,data,...}}) =
	    let
		fun spectree (addedtree, init) =
		    (wout "s comb\n";
		     Itag.Tag.combine(lastspectree, addedtree,
				     Itag.Tag.CI{address=address,
						 words=name,
						 match=match,
						 state=init}))

		fun gentree (addedtree, init) =
		    (wout "g comb\n";
		     Itag.Tag.combine(lastgentree, addedtree,
				      Itag.Tag.CI{address=address,
						  words=name,
						  match=match,
						  state=init}))
			      
		fun zip (wcall, midgen, tcall, init, tree) =
		    case (Itag.Call.match(wcall, tcall, specinfo),
			  Itag.Call.match(wcall, tcall, midgen),
			  spectree (tree, init),
			  gentree (tree, init)) of
			(SOME (_, newspec), 
			 SOME (_, newgen), 
			 SOME spectinst,
			 SOME gentinst) =>
			    let
				val problem = 
				    Constraint.addConstraint 
				    (lastproblem, presup,
				     Constraint.Call.getInit
				     (newgen, parms, #1 public), newgen,
				     Constraint.Call.getInit
				     (newspec, parms, #1 public), newspec)
				val newsemantics =
				    (case semantics of
					 NONE => lastsem
				       | SOME elt => elt :: lastsem)
				val newcgoals =
				    (List.filter
				     (notEntailed (public, sharedscheme, 
						   newspec, newsemantics))
				     lastcgoals)
				val newsizes = Constraint.sizes problem
				val newwork = Constraint.worksizes problem
				val newnumfree = Constraint.numfree problem

				val _ =
				    if Debug.debugp "derive" then
					(print "new problem is: ";
					 Constraint.printProblem problem)
				    else
					()
					
				val issubst = 
				    case Itag.Tag.getTkind tree of
					Itag.Tag.Initial => true
				      | _ => false

				val incompleteness = 
				    Itag.Tag.incompleteness spectinst
			    in
				SOME (DS{
					 constraint = problem,
					 specinfo = newspec,
					 spectinst = spectinst,
					 geninfo = newgen,
					 gentinst = gentinst,
					 semantics = newsemantics,
					 score = mkScore(newcgoals, issubst,
							 basic, newsizes, newwork,
							 newnumfree, incompleteness)
					 })
			    end
		      | (x,y,z,w) => 
			    (let
				 fun s NONE = print "N"
				   | s _ = print "Y"
			     in
				 print "Ack: (derive) zip failed\nstatus: ";
				 s x; s y; s z; s w;
				 print "\ncall: ";
				 Itag.Call.prettyprintCall specinfo call;
				 print "\ntcall: ";
				 Itag.Call.prettyprintCall specinfo tcall;
				 NONE
			     end)

		fun doTree (nil, midgen)  _ = NONE
		  | doTree (call::more, midgen) 
		    (tm as Itag.M{info=tinfo, init=init,
				    call=tcall, tree}) =

		    (if Itag.Call.duplicates(call, specinfo, tcall, tinfo) then
			 (wout "zipping\n";
			  zip (call, midgen, tcall, init, tree))
		     else 
			 doTree(more, midgen) tm)
			
		val _ = 
		    if Debug.debugp "derive" then
			(print "Assembling results for ";
			 app (fn x => (print x; print " ")) name)
		    else ()

		fun listoffirst f nil = nil
		  | listoffirst f (h::r) =
		    case f h of
			SOME v => [v]
		      | NONE => listoffirst f r

		val result =
		    (case Itag.Tag.opAtAddress(lastgentree, address) of
			 NONE => 
			     (print "Ack: (Derive) cannot construct generic match\n";
			      nil)
		       | SOME (Itag.Tag.OP{data=gendata,...}) => 
			     case Itag.Call.match(call, gendata, geninfo) of 
				 NONE =>
				     (print "Ack: (derive) no generic match\n";
				      nil)
			       | SOME (_, midgen) => 
				     if Debug.debugp "allchoice" then
					 (List.mapPartial 
					  (doTree (trees, midgen)) 
					  checkedtrees)
				     else
					 (listoffirst
					  (doTree (trees, midgen)) 
					  checkedtrees))
		val _ =
		    if Debug.debugp "derive" then
			(print "Assembled results for ";
			 app (fn x => (print x; print " ")) name;
			 print ": there are ";
			 print (Int.toString (length result));
			 print "\n")
		    else
			()
	    in
		result
	    end

	fun extensions (words, trees, public, shared, private) 
	    (DS{constraint, 
		specinfo, spectinst,
		geninfo, gentinst, 
		semantics, score = SI{cgoals,...},...}) =
	    let
		fun matchw oppo = 
		    (if Debug.debugp "derive" then
			 (print "Matching words for ";
			  Itag.Tag.printOp oppo;
			  print "\n")
		     else
			 ();
		     Word.matchAll (public, private) (oppo, specinfo) words)

		val oppos = Itag.Tag.choices (spectinst)

		val instWords = foldr op@ nil (map matchw oppos) 
		val instTrees = Itag.matchAll public (Word.getCalls instWords) trees
		    
		val interm = 
		    (foldr op@ nil
		     (map (assemble (public, shared, instTrees, 
				     semantics, constraint,
				     cgoals, geninfo, spectinst, gentinst)) 
		      instWords))

		val final  = quicksort cmpScore interm

		val _ = 
		    if Debug.debugp "comp" then
			(print "before sorting: options are: \n";
			 app printState interm;
			 print "\n\nafter sorting: options are: \n";
			 app printState final)
		    else
			()
	    in
		final
	    end
		   
	datatype generand = 
	    GN of {name: string,
		   private: Constraint.Formula.formula,
		   shared: Constraint.Formula.formula,
		   task: Constraint.Call.call,
		   pattern: Constraint.Call.call,
		   features: Itag.Tag.Features.features,
		   cgoals: Constraint.Formula.formula list}

	  | PGSP of {name: string,
		     private: Constraint.Formula.formula,
		     discourse: Constraint.Formula.formula,
		     shared: Constraint.Formula.formula,
		     extender: Constraint.Formula.formula,
		     taskSchema: Constraint.Call.Term.term,
		     schemaKey: int,
		     pattern: Constraint.Call.call,
		     features: Itag.Tag.Features.features,
		     mainQuery: Constraint.Formula.formula,
		     auxQuery: Constraint.Formula.formula}

	datatype generand_part =
	    NAME of string
	  | FMLA of Constraint.Formula.formula
	  | CALL of Constraint.Call.call
	  | FEATURES of Itag.Tag.Features.features
	  | GOALS of Constraint.Formula.formula list
	  | TERM of Constraint.Call.Term.term
	  | KEY of int
	    
	local
	    open FieldStr Constraint.Formula Constraint.Call
	in
	    fun generandFields collection "gen" =
		[F{name = "name",
		   read = wrapReader(SOME o (NAME o merge), readStrings)},
		 F{name = "private",
		   read = wrapReader(SOME o FMLA, readFormula collection)},
		 F{name = "shared",
		   read = wrapReader(SOME o FMLA, readFormula collection)},
		 F{name = "describe",
		   read = wrapReader(SOME o CALL, readCall (newPPVar ()))},
		 F{name = "pattern",
		   read = wrapReader(SOME o CALL, readCall (newPPVar ()))},
		 F{name = "features",
		   read = wrapReader(SOME o FEATURES, Itag.Tag.Features.readFeatures)},
		 F{name = "communicate",
		   read = wrapReader(SOME o GOALS, readListOf (readFormula collection))}]

	      | generandFields collection "desc" =
		[F{name = "name",
		   read = wrapReader(SOME o (NAME o merge), readStrings)},
		 F{name = "private",
		   read = wrapReader(SOME o FMLA, readFormula collection)},
		 F{name = "discourse",
		   read = wrapReader(SOME o FMLA, readFormula collection)},
		 F{name = "shared",
		   read = wrapReader(SOME o FMLA, readFormula collection)},
		 F{name = "extender",
		   read = wrapReader(SOME o FMLA, readFormula collection)},
		 F{name = "schema",
		   read = wrapReader(SOME o TERM, Term.readTerm)},
		 F{name = "variable",
		   read = fn nil => (NONE, nil) | (s::rest) => (SOME (KEY (Term.getVar s)), rest)},
		 F{name = "pattern",
		   read = wrapReader(SOME o CALL, readCall (newPPVar ()))},
		 F{name = "features",
		   read = wrapReader(SOME o FEATURES, Itag.Tag.Features.readFeatures)},
		 F{name = "hit",
		   read = wrapReader(SOME o FMLA, readFormula collection)},
		 F{name = "communicate",
		   read = wrapReader(SOME o FMLA, readFormula collection)}]

	      | generandFields _ _ = raise Match

	    fun assembleGenerand "gen"
		[NAME s, FMLA private, FMLA shared, 
		 CALL task, CALL pattern, FEATURES features,
		 GOALS goals] =
		SOME (GN{name = s,
			 private = private,
			 shared = shared,
			 task = task,
			 pattern = pattern,
			 features = features,
			 cgoals = breakFormulas goals})
	      | assembleGenerand "desc" 
		[NAME name, FMLA private, FMLA discourse, FMLA shared, FMLA extender,
		 TERM taskSchema, KEY schemaKey, CALL pattern, FEATURES features,
		 FMLA hit, FMLA communicate] =
		SOME (PGSP{name = name,
			   private = private,
			   discourse = discourse,
			   shared = shared,
			   extender = extender,
			   taskSchema = taskSchema,
			   schemaKey = schemaKey,
			   pattern = pattern,
			   features = features,
			   mainQuery = hit,
			   auxQuery = communicate})
	      | assembleGenerand _ _ = NONE

	    fun readGenerand collection (s::more) =
		(case readstruct (s, generandFields collection s) (s::more) of
		     (GOT f, s') => (assembleGenerand s f, s')
		   | (EMPTY, _) => (NONE, s::more)
		   | (ERROR e, _) => (print e; (NONE, s::more)))
	      | readGenerand collection nil = (NONE, nil)

	    fun gfieldInits "gen" =
		[ "hello", "I $", "C $", "s(hello)", "s(X)", "()", "" ]
	      | gfieldInits "desc" =
		[ "hello", "I $", "C $", "C $", "s(X)", "X", "s(Y)", "()", "I $" ]
	      | gfieldInits _ = raise Match
	end

	fun expandAbbrev (database, private, shared, abbrev) =
	    let
		val (world, queryfun) =
		    Constraint.Formula.setof(fn _ => nil, (), private, database)

		val blankInit = 
		    let 
			open Constraint.Call.Term
		    in 
			IntMap.empty () : init IntMap.intmap
		    end

		val proofs = queryfun (blankInit, abbrev)

		val public = 
		    Constraint.Formula.setof(fn _ => nil, (), 
					     (Constraint.Formula.compose 
					      shared private), 
					     database)

		val possibleGoals = 
		    (List.mapPartial 
		     Constraint.Formula.readoutMetavar
		     proofs)

	    in
		List.filter
		(notEntailed (public, shared, 
			      Constraint.Call.mkInfo(), nil))
		possibleGoals
	    end

	fun myGenerateList (words, trees, database)
	    (GN{name, private, shared, task, pattern, features, cgoals}) =
	    let
		open Constraint.Formula 
		val privq = setof (fn _ => nil, (), private, database)
		val pubq = setof (fn _ => nil, (), 
				  compose shared private, database)
		val init = start(features, task, pattern, cgoals, pubq)
		val extend = extensions (words, trees, pubq, shared, privq)

		fun next(point, sofar) =
		    let
			val _ = 		    
			    if Debug.debugp "step" then
				(print "State is: ";
				 printState point;
				 print "\nhit enter to continue\n";
				 while TextIO.inputN(TextIO.stdIn,1) <> "\n" do ())
			    else
				()

			val more = extend point

			fun scan nil =
			    (print "No more improvement\n";
			     (finalize (privq,point)) :: point :: sofar)
			  | scan (hd :: rest) =
			    if completeness hd = SOME 0 then
				(print "Derivation completed successfully\n";
				 (finalize (privq,hd)) :: hd :: point :: sofar)
			    else 
				if cmpIter (hd, point) then
				    next (hd, point :: sofar)
				else
				    scan rest
		    in
			(case more of 
			     nil => 
				 (print "No further actions possible\n";
				  if Debug.debugp "step" then
				      (print "State is: ";
				       printState point;
				       print "\nState finalizes as: ";
				       printState (finalize (privq,point));
				       print "\nhit enter to continue\n";
				       while TextIO.inputN(TextIO.stdIn,1) <> "\n" do ())
				  else
				      ();
				  (finalize (privq,point)) :: point :: sofar)
			   | opts => 
				 scan opts)
		    end

	    in
		rev (next (init,nil))
	    end

	  | myGenerateList (words, trees, database)
	    (PGSP{name, private, discourse, shared, extender,
		  taskSchema, schemaKey, pattern, features, 
		  mainQuery, auxQuery}) =
	    let
		val gen = myGenerateList (words, trees, database)
		val (world, queryfun) = 
		    Constraint.Formula.setof(fn _ => nil, (), private, database)
		fun last nil = raise Empty
		  | last [one] = one
		  | last (x::r) = last r

		fun continueParagraph (generand as GN{private, cgoals,...}, 
				       maingoals, 
				       sofar) =
		    let
			val nextSentence as 
			    DS{specinfo, semantics, spectinst, 
			       score=SI{cgoals=remainingCgoals,...},...} = 
			    last (gen generand)
			val newprivateSkeleton =
			    Constraint.Formula.substImplic(semantics, 
							   private, 
							   discourse,
							   extender)
			val newprivate =
			    Constraint.Formula.subst
			    (Constraint.Call.getKeyInit 
			     (specinfo, Constraint.Formula.freeVars newprivateSkeleton, world),
			     world)
			    (newprivateSkeleton)

			val remainingMaingoals = 
			    (List.filter
			     (notEntailed 
			      (Constraint.Formula.setof(fn _ => nil, (), 
							Constraint.Formula.compose shared newprivate, 
							database), 
			       shared, Constraint.Call.mkInfo(), nil))
			      maingoals)
		    in
			if null remainingMaingoals then
			    (print "Story essentially successful\n";
			     rev (nextSentence :: sofar))
			else 
			    if (length remainingCgoals) = (length cgoals) then
				(print "Story terminally unsuccessful\n";
				 print "Length of remainingCgoals is ";
				 print (Int.toString (length remainingCgoals));
				 print "\nLength of cgoals is ";
				 print (Int.toString (length cgoals));
				 print "\nThis is the final context:\n";
				 Constraint.Formula.printFormula print newprivate;
				 print "\nThese are the remaining main goals:\n";
				 app (fn x => (Constraint.Formula.printFormula print x;
					       print "\n")) remainingMaingoals;
				 rev sofar)
			    else 
				let
				    val newTask =
					let
					    open Constraint.Formula Constraint.Call
					in
					    createCall(Term.subst(bindArg(hd remainingMaingoals, 
									  schemaKey, world))
						       taskSchema)
					end
				in
				    continueParagraph(GN{name = name,
							 private = newprivate,
							 shared = shared,
							 task = newTask,
							 pattern = pattern,
							 features = features,
							 cgoals = remainingCgoals},
						      remainingMaingoals,
						      nextSentence :: sofar)
				end
		    end
		  | continueParagraph _ = raise Match

		val maingoals = 
		    expandAbbrev(database, private, shared, mainQuery) 
		    
		val allgoals = 
		    maingoals @
		    expandAbbrev(database, private, shared, auxQuery)

		val _ = 
		    if Debug.debugp "derive" then
			(print "Will consider the following story goals:\n";
			 app (fn x => (Constraint.Formula.printFormula print x;
				       print ".\n")) 
			 allgoals)
		    else ()

	    in
		case maingoals of 
		    nil => nil
		  | (head :: _) => 
			let
			    val newTask = 
				let
				    open Constraint.Formula Constraint.Call
				in
				    createCall(Term.subst(bindArg(head, schemaKey, world))
					       taskSchema)
				end
			in
			    continueParagraph(GN{name = name,
						 private = private,
						 shared = shared,
						 task = newTask,
						 pattern = pattern,
						 features = features,
						 cgoals = allgoals}, 
					      maingoals,
					      nil)
			end
	    end

	fun generateList info eg =
	    Clock.clockapp (myGenerateList info) 
	    "overall generation task"
	    eg

	val _ = app Debug.register 
	    [("dosubst",
	      "Never stop with failure if a substitution is possible\n",
	      true),
	     ("derive", 
	      "Describe what's happening in word choice\n", 
	      false),
	     ("comp",
	      "Describe comparison between different alternatives\n",
	      false),
	     ("allchoice",
	      "Evaluate different syntactic choices separately\n",
	      false),
	     ("step",
	      "Wait for return after selecting next move\n",
	      false)]
    end
