
functor TagCall (structure Call : CALL;
		 structure Tag : TAG ;
		 structure Formula : FDB;
		 sharing Call.Term = Formula.Term;
		 sharing type Tag.data = Call.call;
		 sharing type Tag.datastate = Call.init) : ITAG =
    struct
	structure Call = Call
	structure Tag = Tag
	structure Formula = Formula

	datatype entry =
	    T of {call: Call.call,
		  tree: Tag.tree,
		  prags: Formula.formula}

	datatype taglex = 
	    L of entry list

	datatype match =
	    M of {info: Call.info,
		  call: Call.call,
		  init: Call.init,
		  tree: Tag.tree}

	datatype entry_part =
	    CALL of Call.call
	  | PRAGS of Formula.formula
	  | RAWTREE of string list

	local 
	    open FieldStr

	in
	    val treePPVar = Call.newPPVar ()

	    val treeDataField = 
		F{name = "type",
		  read = Call.readCall treePPVar}

	    fun entryFields (collection) =
		[F{name = "name",
		   read = wrapReader(SOME o CALL, Call.readCall treePPVar)},
		 F{name = "pragmatics",
		   read = wrapReader(SOME o PRAGS, Formula.readFormula collection)}]

	    val treeField = F{name = "tree", read = readStrings}
	    val treeEntryField = F{name = "tree", read = wrapReader(SOME o RAWTREE, readStrings)}

	    val fieldTreeInit = (FieldStr.Tree{key = "new tree",
					       data = ("anchor", ["1"]),
					       kids = nil},
				 [ "t", "true" ])

	    fun readEntry collection s =
		let
		    val fields = entryFields (collection) @ [treeEntryField]

		    fun assemble ([CALL call, PRAGS prags, RAWTREE ts]) =
			(case Tag.readTree treeDataField ts of
			     (SOME subtree, nil) =>
				 (SOME (T{call = call,
					  tree = subtree,
					  prags = prags}))
			   | (SOME _, s) =>
				 (print "Tree field not all consumed\n";
				  print "tree field is:";
				  print (merge ts);
				  print "\nremainder is ";
				  print (merge s);
				  print "\n";
				  (NONE))
			   | (NONE, s) => NONE)
		      | assemble _ =
			(print "Mismatched fields!\n";
			 (NONE))
		in
		    case readstruct("entry", fields) s of
			(GOT a, s') => (assemble a, s')
		      | (EMPTY, s') => (NONE, s)
		      | (ERROR e, s') => (print e;
					  (NONE, s))
		end

	    fun writerecEntry (tree, info) =
		let
		    fun writedata c = (Call.writeHead c, Call.writeprettyCall info c)
		in
		    Tag.writerecTree (writedata) tree
		end
	end

	fun newTaglex () =  L nil

	fun addEntry (L el, e) = L(e :: el)

	fun printEntry (T{call, tree, prags}) =
	    (Call.printCall call;
	     print " - ";
	     Tag.printTreeLeaves tree;
	     print " - ";
	     Formula.printFormula print prags;
	     print "\n")
	     
	fun printTaglex (L entries) = 
	    app printEntry entries

	fun printMatch (M{info, call, tree, ...}) =
	    (Call.prettyprintCall info call;
	     print " - ";
	     Tag.printTreeLeaves tree;
	     print "\n")

	fun matchEntry (publicworld, publicset) 
	    (pattern, info) 
	    (T{call, tree, prags}) =
	    let
		val (newcall, newinit) = Call.newcopyCall(call, Call.mkInit())

		val newinfo = Call.match(pattern, newcall, info)

		fun uniquecxt (i) =
		    (case publicset (Call.getInit(i, newcall, publicworld), 
				     prags) of
			 [one] => true
		       | _ => false)
	    in
		case newinfo of 
		    NONE => NONE
		  | SOME (_, i) => 
			if uniquecxt i then
			    SOME (M{info = i, 
				    init = newinit,
				    call = newcall,
				    tree = tree})
			else
			    NONE
	    end

	fun matchAll logic (calls) (L entries) =
	    let
		fun checknext (nil, _, _, results) = rev results
		  | checknext (entry::rest, nil, failed, results) = 
		    checknext (rest, failed, nil, results)
		  | checknext (entry::rest, some::more, failed, results) =
		    case matchEntry logic some entry of
			NONE => 
			    checknext(entry::rest, more, some::failed, results)
		      | SOME result =>
			    checknext(entry::rest, more, failed, result :: results)
	    in
		checknext(entries, calls, nil, nil)
	    end

	fun specOrder test (L entries) =
	    let
		open Formula
		fun asSpec (T{prags=p1,...}, T{prags=p2,...}) =
		    case test (Term.IntMap.empty(), mkClosedImplic([p1],p2)) of
			nil => 
			    (false)
		      | _ => 
			    (true)

		fun findleast (t, nil, left) = (t, rev left)
		  | findleast (t, t'::m, left) =
		    if asSpec(t', t) then
			findleast(t', m, t :: left)
		    else
			findleast(t, m, t' :: left)

		fun sort (nil, done) = rev done
		  | sort (t::more, done) =
		    let
			val (t', next) = findleast(t,more,nil)
		    in
			sort(next, t'::done)
		    end

		val result = sort(entries, nil)
		val _ =
		    if Debug.debugp "trees" then
			(print "Entries initially:\n";
			 app (fn (T{call,...}) => (Call.printCall call; print ";\n")) entries;
			 print "\nSorted entries:\n";
			 app (fn (T{call,...}) => (Call.printCall call; print ";\n")) result;
			 print "\n")
		    else
			()
			      
	    in
		L result
	    end

	val _ =
	    Debug.register ("trees", "Print stuff about tree database\n", false)
    end
