
functor Tag(structure Features : FEATURES;
	    type data;
	    type datastate;
	    val newcopyData : data * datastate -> data * datastate) 
    : TAG =
    struct
	structure Features = Features

	open Features

	type data = data
	type datastate = datastate

	datatype tnode =
	    Leaf of string
	  | Anchor of {index:int, top:features}
	  | Subst of {top: features}
	  | Internal of {top: features,
			 bottom: features,
			 kids: tnode list}

	type address = int list

	datatype kind = 
	    ToSubst
	  | ToAdjoin

	datatype opportunity = 
	    OP of {kind: kind,
		   address: address,
		   match: features,
		   data: data}

	datatype cinfo =
	    CI of {address: address,
		   words: string list,
		   match: Features.features,
		   state: datastate}

	datatype tkind =
	    Initial 
	  | Auxiliary of address
	    
	datatype tree =
	    T of {kind: tkind,
		  stuff: tnode,
		  match: features,
		  handles: opportunity list}

	datatype tinst =
	    I of {stuff: tnode,
		  binding: binding,
		  handles: opportunity list}

	exception TagError of string
	exception failure

	fun getTkind (T{kind,...}) = kind

	fun headFeatures (Leaf _) = topFeature ()
	  | headFeatures (Anchor {top,...}) = top
	  | headFeatures (Subst {top,...}) = top
	  | headFeatures (Internal {top,...}) = top

	fun promoteFeatures (t as Leaf _) = t
	  | promoteFeatures (Anchor {top,index}) = 
	    Anchor{top=topFeature(),index=index}
	  | promoteFeatures (Subst{top}) = 
	    Subst{top=topFeature ()}
	  | promoteFeatures (Internal{top,bottom,kids}) = 
	    Internal{top=bottom, bottom=bottom, kids=kids}
  
	fun setHeadFeatures (t as Leaf _, f) = t
	  | setHeadFeatures (t as Anchor {index,top}, f) = 
	    Anchor{index = index, top = f}
	  | setHeadFeatures (Subst {top}, f) = Subst{top=f}
	  | setHeadFeatures (Internal {top,bottom,kids}, f) =
	    Internal{top=f, bottom=bottom, kids=kids}

	fun copyTNode lex (Subst {top}, (binding, headfeatures)) = 
	    let
		val (copy, newbinding) = newcopyFeatures (top, binding)
	    in
		(Subst{top=copy}, (newbinding, headfeatures))
	    end
	  | copyTNode lex (Internal{top,bottom,kids}, (binding, headfeatures)) =
	    let
		val (copytop, binding1) = newcopyFeatures(top, binding)
		val (copybottom, binding2) = newcopyFeatures(bottom, binding1)
		val (copykids, (binding3, newheadfeatures)) = 
		    U.foldstate (copyTNode lex) (kids, (binding2, headfeatures))
	    in
		(Internal{top=copytop,
			  bottom=copybottom,
			  kids=copykids},
		 (binding3,
		  newheadfeatures))
	    end
	  | copyTNode (words) (Anchor {index=1,top}, (binding, NONE)) = 
	    let
		val (copytop, newbinding) = newcopyFeatures(top, binding)
	    in 
		(Leaf (hd(words) handle Hd => "????"), (newbinding, SOME copytop))
	    end
	  | copyTNode (words) (Anchor {index=n,top}, (binding, headfeatures)) =
	    (Leaf (List.nth(words,n-1) handle Subscript => "????"), (binding, headfeatures))

	  | copyTNode lex s = s

	fun printAddress (nil) = 
	    (print "<root>")
	  | printAddress (address:int list) = 
	    app (fn x => (print (Int.toString x);
			  print " ")) address

	local
	    fun printKind (ToSubst) = print "alpha"
	      | printKind (ToAdjoin) = print "beta"
	in
	    fun printOp (OP{kind,address,...}) = 
		(print "OP(";
		 printAddress address;
		 print ": ";
		 printKind kind;
		 print ")")
	end
			      

	local 
	    fun printTNode (pf : features -> unit) =
		let
		    (* indent n: insert n spaces *)
		    fun indent 0 = ()
		      | indent n = 
			(print " ";
			 indent (n - 1))

		    (*
		     * pt -- short for print tree -- takes integer
		     * number of spaces to indent and tree and shows
		     * it in a form rather like what could be read
		     * back in.  pcat is used to show syntactic
		     * features that label trees. 
		     *)
		    fun pt il (Leaf s) = 
			(indent il;
			 print s;
			 print "\n")

		      | pt il (Anchor {index=n,...}) =
			(indent il;
			 print "<#";
			 print (Int.toString n);
			 print ">\n")

		      | pt il (Subst{top}) =
			(indent il;
			 print "[ subst top: ";
			 pf top;
			 print " ]\n")
			
		      | pt il (Internal{top, bottom, kids}) =
			(indent il;
			 print "[ node top: ";
			 pf top;
			 print "\n";
			 indent (il + 6);
			 print "bottom: ";
			 pf bottom;
			 print "\n";
			 app (pt (il + 3)) kids;
			 indent (il + 2);
			 print "]\n")
		in
		    pt 0
		end
	in
	    val dumpTNode = printTNode (printFeatures)
	    fun prettyprintTNode binding = printTNode (prettyprintFeatures binding)
	end

	local 
	    fun printkind (Initial) = 
		(print "initial tree")
	      | printkind (Auxiliary _) =
		(print "auxiliary tree")

	    fun printleaves (outstring,bind,features) (Leaf s) =
		outstring(s, features, bind)

	      | printleaves (_) (Anchor {index=i,...}) =
		(print "<#";
		 print (Int.toString i);
		 print "> ")

	      | printleaves (outstring, bind, _) (Internal{bottom,kids,...}) =
		(app (printleaves (outstring, bind, bottom)) kids)

	      | printleaves (_) (Subst _) =
		(print "_ ")

	    fun printTreeAux outstring (T{kind,stuff,...}) =
		(print "Type of ";
		 printkind kind;
		 print ": ";
		 printleaves (outstring, emptyBinding(), topFeature ()) stuff;
		 print "\n")

	    fun plain (s:string, _, _) =
		(print s;
		 print " ")

	    fun fancy (lookup) (s:string, f, b) =
		case lookup (s, f, b) of
		    NONE => 
			(print " /";
			 print s;
			 print "/")
		  | SOME (s':string) =>
			(case s' of
			     "" => ()
			   | _ => 
				 (if Char.isPunct (String.sub(s',0)) then
				      print s'
				  else
				      print (" " ^ s')))
	in
	    val printTreeLeaves = printTreeAux plain
	    fun prettyprintTreeLeaves f = printTreeAux (fancy f)
	end

        local 
	    datatype lastEltType =
		NoLastType
	      | StickyPunctuationType
	      | OtherType

	    fun upcase (s) =
		case explode s of
		    nil => s
		  | a::r => implode(Char.toUpper a :: r)

	    fun surround(s,false) = s
	      | surround(s,true) = "/" ^ s ^ "/"

	    fun issticky (s) =
		case rev(explode s) of
		    #"-" :: _ => true
		  | #"(" :: _ => true
		  | #"`" :: _ => true
		  | _ => false
	
	    fun realize (s, NoLastType, stemp) =
		if s = "" then
		    (surround(s,stemp), NoLastType)
		else 
		    if not (Char.isPunct (String.sub(s, 0))) orelse 
			(ord (String.sub(s,0)) = ord #"_") then
			(surround(upcase s,stemp), OtherType)
		    else
			(surround(s,stemp), NoLastType)


	      | realize (s, StickyPunctuationType, stemp) =
		(surround(s,stemp),
		 if issticky s orelse s = "" then StickyPunctuationType else OtherType)

	      | realize (s, OtherType, stemp) =
		let
		    val next =
			if issticky s then StickyPunctuationType
			else OtherType
		in
		    case explode s of
			#"(" :: _ => (" " ^ (surround(s,stemp)), next)
		      | #"<" :: _ => (" " ^ (surround(s,stemp)), next)
		      | #"`" :: _ => (" " ^ (surround(s,stemp)), next)
		      | #"_" :: _ => (" " ^ (surround(s,stemp)), next)
		      | nil => ("", next)
		      | h :: _ =>
			    if Char.isPunct(h) then
				(surround(s,stemp), next)
			    else
				(" " ^ (surround(s,stemp)), next)
		end

	    fun plain (s:string, _, _, _) = (s ^ " ", OtherType)

	    fun fancy (lookup) (s:string, f, b, elttype) =
		case lookup (s, f, b) of 
		    NONE => realize(s,elttype,true)
		  | SOME (s2:string) => realize(s2,elttype,false)

	    fun writeleaves (writestring,bind,features) (Leaf s, (sl,elttype)) =
		let
		    val (w,n) = writestring(s, features, bind, elttype) 
		in
		    (w :: sl, n)
		end

	      | writeleaves (_) (Anchor {index=i,...}, (sl,elttype)) = 
		let
		    val (w,n) = realize("<#" ^ (Int.toString i) ^ ">",elttype,false)
		in 
		    (w :: sl, n)
		end 

	      | writeleaves (writestring, bind, _) (Internal{bottom,kids,...}, t) =
		(foldl (writeleaves (writestring, bind, bottom)) t kids)

	      | writeleaves (_) (Subst _, (sl,elttype)) = 
		let
		    val (w,n) = realize("_",elttype,false)
		in
		    (w :: sl, n)
		end

	    fun writeTinstAux writestring (I{stuff,binding,...}) =
		String.concat
		(rev (#1 (writeleaves 
			  (writestring, binding, topFeature ()) 
			  (stuff,(nil,NoLastType)))))

	in
	    val writeTinstLeaves = writeTinstAux plain
	    fun writeprettyTinstLeaves f = writeTinstAux (fancy f)
		
	    val printTinstLeaves = print o writeTinstLeaves
	    fun prettyprintTinstLeaves f = print o (writeprettyTinstLeaves f)
	end

	datatype tree_part =
	    NAME of string list
	  | DATA of data
	  | FEATURE of features
	  | INDEX of int
	  | REC of string * tree_part list
	  | KIDS of tree_part list

	val mkData = DATA

	local
	    open FieldStr
	    exception UnknownType

	    fun nextCt (NONE) = NONE
	      | nextCt (SOME ct) = (SOME (ct+1))

	    fun subCt (NONE, address) = address
	      | subCt (SOME ct, address) = ct :: address

	    fun mkAddr guy = rev (subCt guy)

	    fun buildTree address 
		(REC("node", [DATA data, FEATURE top, FEATURE bot, KIDS kids]), (ops, foot, ct)) =
		let
		    val (kidtrees, (newops, nfoot, _)) =
			U.foldstate (buildTree (subCt(ct, address))) 
			(kids, (ops, foot, SOME 0))
		in
		    (Internal{top = top, bottom = bot, kids = kidtrees}, 
		     (OP{kind = ToAdjoin, address = mkAddr(ct, address),
			 match = top, data = data} :: newops,
		      nfoot,
		      nextCt ct))
		end
	      | buildTree address
		(REC("foot", [DATA data, FEATURE top]), (ops, foot, ct)) =
		(Subst{top=top}, (ops, SOME (mkAddr(ct, address)), nextCt ct))
	      | buildTree address 
		(REC("subst", [DATA data, FEATURE top]), (ops, foot, ct)) =
		(Subst{top=top}, 
		 (OP{kind = ToSubst, address = mkAddr(ct, address),
		     match = top, data = data} :: ops,
		  foot, 
		  nextCt ct))
	      | buildTree address
		(REC("anchor", [INDEX index]), (ops, foot, ct)) =
		(Anchor {index=index, top=topFeature ()}, 
		 (ops, foot, nextCt ct))		
	      | buildTree address
		(REC("head", [FEATURE top]), (ops, foot, ct)) =
		(Anchor {index=1, top=top}, 
		 (ops, foot, nextCt ct))
	      | buildTree address
		(REC("words", [NAME strings]), (ops, foot, ct)) =
		 (Leaf (merge strings), 
		  (ops, foot, nextCt ct))
	      | buildTree _ _ =
		 raise UnknownType
	in
	    fun treeFields (dataField as F{name, read}) =
		fn "node" => ([F{name = name,
				 read = wrapReader (SOME o DATA, read)},
			       F{name = "top", 
				 read = wrapReader (SOME o FEATURE, readFeatures)},
			       F{name = "bottom",
				 read = wrapReader (SOME o FEATURE, readFeatures)}],
		[F{name = "kids",
		   read = wrapReader (SOME o KIDS, 
				      readListOf (composeTree dataField))}])

		 | "foot" => ([F{name = name,
				 read = wrapReader (SOME o DATA, read)},
			       F{name = "top", 
				 read = wrapReader (SOME o FEATURE, readFeatures)}],
			      nil)
		       
		 | "subst" => ([F{name = name,
				  read = wrapReader (SOME o DATA, read)},
				F{name = "top", 
				  read = wrapReader (SOME o FEATURE, readFeatures)}],
			       nil)

		 | "anchor" => ([F{name = "index",
				   read = wrapReader (SOME o INDEX, readInt)}],
				nil)

		 | "head" => ([F{name = "features",
				   read = wrapReader (SOME o FEATURE, readFeatures)}],
				nil)

		 | "words" => ([F{name = "words",
				  read = wrapReader (SOME o NAME, readStrings)}],
			       nil)
		 | _ => raise UnknownType

	    and composeTree (dataField) nil = (NONE, nil)
	      | composeTree (dataField) (s :: more) =
		let
		    val fields = op@ (treeFields dataField s)
		in
		    (case readstruct (s, fields) (s :: more) of
			 (GOT a, s') => (SOME (REC (s,a)), s')
		       | (EMPTY, s') => (NONE, s')
		       | (ERROR e, s') =>
			     (print e; (NONE, s')))
			 
		end  
	    handle UnknownType => (NONE, s :: more)

	    fun readTree dataField s =
		(case composeTree dataField s of
		     (NONE, s') => (NONE, s')
		   | (SOME stuff, s') => 
			 let
			     val (tnode, (ops, footo, _)) = 
				 buildTree nil (stuff, (nil, NONE, NONE))
			 in
			     (SOME (T{kind = (case footo of
						  NONE => Initial
						| SOME address =>
						      Auxiliary address),
				      stuff = tnode,
				      match = headFeatures tnode,
				      handles = ops}),
			     s')
			 end) handle UnknownType => (NONE, s)

	    fun segmentTree dataField (s::more) =
		(let
		     val (flat, recursive) = (treeFields dataField s)
			 
		     fun kidf (nil, sofar) = sofar
		       | kidf (s, sofar) =
			 (case segmentTree dataField s of
			      (GOT a, s') => kidf(s', a :: sofar)
			    | (EMPTY, _) => rev sofar
			    | _ => raise UnknownType)
		 in
		     case rawSegment (s, flat @ recursive) (s::more) of
			 (GOT a, s') =>
			     let 
				 val len = List.length flat
				 val flatf = List.take (a, len)
				 val recf = List.drop (a, len)
				 fun extractName (_ :: ":" :: s :: _) = s
				   | extractName (s :: _) = s
				   | extractName _ = "empty"
			     in
				 (GOT (Tree{key = extractName (hd flatf) handle Hd => "empty",
					    data = (s, map merge flatf),
					    kids = rev (foldl kidf nil recf)}),
				     s')
			     end
		       | (EMPTY, s) => (EMPTY, s)
		       | (ERROR e, s) => (ERROR e, s)
		 end
	     handle UnknownType => (ERROR ("unknown tree type " ^ s ^ ".\n"), s::more))
	      | segmentTree _ nil = (EMPTY, nil)

	    fun writeForTree datafield (key, flatargs, recargs) =
		(case treeFields datafield key of
		     (flat, nil) => 
			 unsegment(key, flat) flatargs
		   | (flat, head :: _) =>
			 unsegment(key, flat @ [head])  (flatargs @ [ merge recargs ]))
		     handle UnknownType => 
			 (print ("Error: (tag.sml) unknown tree type " ^ key ^ "\n");
			  "")				 
	end

	fun writerecTree writedata (I{stuff, binding, handles}) =
	    let
		fun df (ct, addr) =
		    let
			val a' = rev (case ct of NONE => addr | SOME c => (c :: addr))
		    in
			case List.find (fn (OP{address,...}) => address = a') handles of
			    NONE => ("<>", "<no operation possible>")
			  | SOME (OP{data,...}) => writedata data
		    end
		fun na (NONE, addr) = addr
		  | na (SOME x, addr) = x :: addr
		fun ic (NONE) = NONE
		  | ic (SOME ct) = SOME(ct+1)
		    
		fun wrtn address (Internal{top, bottom, kids}, ct) =
		    let
			val (name, datastr) = df (ct, address)
		    in
			(FieldStr.Tree{key = name,
				       data = ("node", [datastr, 
							writeprettyFeatures binding top, 
							writeprettyFeatures binding bottom]),
				       kids = #1 (U.foldstate (wrtn (na (ct, address))) (kids, SOME 0)) },
			ic ct)
		    end
		  | wrtn address (Subst{top}, ct) =
		    let
			val (name, datastr) = df (ct, address)
		    in
			(FieldStr.Tree{key = name,
				       data = ("subst", [datastr,
							 writeprettyFeatures binding top]),
				       kids = nil},
			 ic ct)
		    end
		  | wrtn address (Anchor {index,...}, ct) =
		    (FieldStr.Tree{key = Int.toString index,
				   data = ("anchor", [ "#" ^ Int.toString index ]),
				   kids = nil},
		     ic ct)
		  | wrtn address (Leaf s, ct) =
		    (FieldStr.Tree{key = s,
				   data = ("words", [s]),
				   kids = nil},
		     ic ct)
	    in
		#1 (wrtn nil (stuff, NONE))
	    end

	local
	    fun substInstReaddress (extension) 
		(OP{kind, address, match, data}, (fcopy, dcopy)) =
		let
		    val (newmatch, newfcopy) = newcopyFeatures(match, fcopy)
		    val (newdata, newdcopy) = newcopyData(data, dcopy)
		in
		    (OP{kind = kind,
			address = extension @ address,
			match = newmatch,
			data = newdata},
		     (newfcopy, newdcopy))
		end

	    fun readdressOp (prefix, extension) =
		let
		    exception mismatch
		    exception kill

		    fun readdress (nil, nil) = raise kill
		      | readdress (nil, address) = extension @ address
		      | readdress (prefix, nil) = raise mismatch
		      | readdress (p :: prefix, a::address) =
			if p = a then
			    p :: readdress(prefix, address)
			else
			    raise mismatch
		in
		    fn old as OP{kind, address, match, data} =>
		    SOME (OP{kind=kind,
			     address=readdress(prefix,address),
			     match=match,
			     data=data})
		    handle mismatch => SOME old
			 | kill => NONE
		end

	    fun different site (opp as OP{address, ...}) =
		if site = address then
		    NONE
		else
		    SOME opp

	    fun modTreeAtAddress modf =
		let
		    fun replace (0, addr, h::t) = 
			let val (newguy, result) = nail (addr, h)
			in (newguy :: t, result)
			end
		      | replace (n, addr, h::t) = 
			let val (newlist, result) = replace (n-1, addr, t)
			in (h :: newlist, result)
			end
		      | replace _ = raise TagError ("Invalid Address: no such node")

		    and nail (nil, tnode) = modf tnode
		      | nail (n::more, Internal{kids,top,bottom}) =
			let val (newkids, result) = 
			    replace (n, more, kids)
			in (Internal{top=top,
				     bottom=bottom,
				     kids= newkids},
			    result)
			end
		      | nail _ = raise TagError ("Invalid Address: wrong type of node")
		in
		    fn x => nail x
		    handle TagError s =>
			("Modification failed; address is ";
			 printAddress (#1 x);
			 print "\nfor site in tree:\n";
			 dumpTNode (#2 x);
			 print "\nproblem is ";
			 print s;
			 print "\n";
			 raise (TagError s))
		end

	    fun substMod (into, xf, binding) tree =
		let 
		    val top = headFeatures tree
		    val itop = headFeatures into
		in
		    case unifyFeatures(top, itop, binding) of
			NONE => raise failure
		      | SOME (newtop, newbinding) => 
			    case unifyFeatures(xf, newtop, newbinding) of
				NONE => raise failure
			      | SOME (ftop, fbinding) => 
				    (setHeadFeatures (into, ftop),
				     fbinding)
		end

	    fun adjoinMod (adjoined, xf, footaddr, binding) tree =
		let
		    val top = headFeatures tree
		    val itop = headFeatures adjoined
		in
		    case unifyFeatures(top, itop, binding) of
			NONE => raise failure
		      | SOME (newtop, newbinding) => 
			    case unifyFeatures (newtop, xf, newbinding) of
				NONE => raise failure
			      | SOME (ftop, fbinding) => 
				    (modTreeAtAddress 
				     (substMod (promoteFeatures tree, 
						topFeature (), fbinding))
				     (footaddr, setHeadFeatures (adjoined, ftop)))
		end
	in
	    fun combine (I{stuff=itree, binding=b1, handles=h1},
			 T{kind=Initial, stuff=goesinto, handles=h2,...},
			 CI{address, words, match, state}) =
		(let
		    val (substituted, (copybind, headfeatures)) = 
			 copyTNode words (goesinto, (emptyBinding(), NONE))

		    val (newtree, newbinding) =
			modTreeAtAddress (substMod (substituted, topFeature (), b1))
			(address, itree)

		    val finalbinding =
			(case headfeatures of 
			     NONE => newbinding
			   | SOME features => 
				 case unifyFeatures(match, features, newbinding) of
				     NONE => raise failure
				   | SOME(_,finalbinding) => finalbinding)

		    val _ = 
			if Debug.debugp "tags" then
			    (print "substituted!\n";
			     prettyprintTNode newbinding newtree;
			     print "\n")
			else
			    ()

		    val (newops, _) =
			U.foldstate (substInstReaddress address)
			(h2, (copybind, state))

		 in
		     SOME (I{stuff=newtree,
			     binding = finalbinding,
			     handles = ((List.mapPartial (different address) h1) @ 
					newops)})
		 end
	     handle failure => NONE)

	      | combine (I{stuff=itree, binding=b1, handles=h1},
			 T{kind=Auxiliary(a2), stuff=around, 
			   handles=h2,...},
			 CI{address, words, match, state}) =
		(let
		     val (adjoined, (copybind, headfeatures)) =
			(copyTNode (words) (around, (emptyBinding(), NONE)))

		     val (newtree, newbinding) =
			 (modTreeAtAddress (adjoinMod (adjoined, topFeature(), a2, b1))
			  (address, itree))
			 
		    val finalbinding =
			(case headfeatures of 
			     NONE => newbinding
			   | SOME features => 
				 case unifyFeatures(match, features, newbinding) of
				     NONE => raise failure
				   | SOME(_,finalbinding) => finalbinding)

		    val _ = 
			if Debug.debugp "tags" then
			    (print "adjoined!\n";
			     prettyprintTNode newbinding newtree;
			     print "\n")
			else
			    ()

		     val (newops, _) = 
			 (U.foldstate (substInstReaddress address)
			  (h2, (copybind, state)))
		 in
		     SOME(I{stuff=newtree,
			    binding=finalbinding,
			    handles= ((List.mapPartial (readdressOp 
							(address, a2)) h1) @ newops)})
		 end
	     handle failure => NONE)
	end

	local
	    exception Failed

	    fun unifytopbot(Subst _, binding) = raise Failed
	      | unifytopbot(Internal{top,bottom,kids}, binding) =
		(case unifyFeatures(top,bottom,binding) of
		     NONE => raise Failed
		   | SOME (f,newbinding) => 
			 let 
			     val (newkids, finalbinding) =
				 U.foldstate unifytopbot (kids, newbinding)
			 in
			     (Internal{top=f,
				       bottom=f,
				       kids=newkids},
			      finalbinding)
			 end)
	      | unifytopbot x = x
	in		     
	    fun finalize (I{stuff, binding, handles}) =
		let
		    val (newtree, newbinding) = 
			unifytopbot(stuff, binding)

		    val _ = 
			if Debug.debugp "tags" then
			    (print "unified top and bottom!\n";
			     prettyprintTNode newbinding newtree;
			     print "\n")
			else
			    ()
		in
		    SOME (I{stuff = newtree,
			    binding = newbinding,
			    handles = nil})
		end
	    handle Failed => NONE
	end

	fun incompleteness (I{stuff, binding, handles}) =
	    let
		fun unifyct (top,bottom) = 
		    case unifyFeatures (top,bottom,binding) of
			NONE => 1
		      | _ => 0

		fun badness (Leaf _, i) = i
		  | badness (Anchor _, i) = i
		  | badness (Subst _, i) = i+1
		  | badness (Internal{top,bottom,kids}, i) =
		    foldl badness (i + (unifyct (top,bottom))) kids 
	    in
		badness (stuff, 0)
	    end

	fun start (features, data) =
	    let
		val (copy, _) = newcopyFeatures (features, emptyBinding ())
	    in
		I{stuff= Subst{top=copy},
		  binding = emptyBinding (),
		  handles = [ OP{kind= ToSubst,
				 address= nil,
				 match = features,
				 data = data} ]}
	    end
	    
	fun choices (I{stuff, binding, handles}) = handles

	fun opAtAddress(I{handles,...}, myaddress) =
	    List.find (fn OP{address,...} => myaddress = address) handles

	val _ = Debug.register("tags", "Describe tree operations\n", false)
    end
