(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 * fdb-sig.sml
 *
 * Signature file for uniform proof search in a fragment of
 * modal logic.
 *)

signature FDB =
    sig
	structure Term : TERM
	structure Patheq : PATHEQ
	sharing Term.Patheq = Patheq

	type formula
	type database
	type 'a proof
	type 'a more

	(* 
	 * read & print out formulas involving modal operators
	 * specified by COLLECTION.  Syntax may vary by
	 * implementation (see 'fdb.sml' for details here).
	 *)
	val readFormula : (Patheq.Modspec.collection -> 
			   string list -> 
			   formula option * string list)
	val readAtom : (string list ->
			formula option * string list)
	val breakFormulas: formula list -> formula list

	val writeFormula : formula -> string
	val printFormula : (string -> unit) -> formula -> unit

	val varFreeIn : int -> formula -> bool
	val freeVars : formula -> Term.IntSet.intset
	val mkImplic : formula list * formula -> formula

	val emptyCxt : formula
	val compose : formula -> formula -> formula
	val substImplic : formula list * formula * formula * formula -> formula
	val mkClosedImplic : formula list * formula -> formula

	(* 
	 * maintain a collection of formulas to use as
	 * axioms in proof search.
	 *)
	val makeDatabase : (Patheq.Modspec.collection *
			    (string * formula) list) -> database
	val addFormulas : database * (string * formula) list -> database

	val makeNewTrans: 
	    ('a proof * Term.init Term.IntMap.intmap * Patheq.pathpt ->
	     Term.init Term.IntMap.intmap)
	val bestAlias: ('a proof * int -> int)

	val rename:
	    Term.init Term.IntMap.intmap * Patheq.pathpt -> formula -> formula
	val subst:
	    Term.init Term.IntMap.intmap * Patheq.pathpt -> formula -> formula

	(*
	 * is the formula true, given assumptions in database.
	 * (for the future: might want to extend this to
	 * report a 'state set' rather than just the bool;
	 * for now, this prints out true substitutions found...)
	 *)
	val queryFormula : 
	    (Term.term * Term.init Term.IntMap.intmap * Patheq.pathpt *
	     'a * Term.system * Patheq.problem -> 
	     ('a * Term.system * Patheq.problem) list) * 'a  ->
	    formula * database -> ('a proof * 'a more) option

	val queryFormulaTr : 
	    (Term.term * Term.init Term.IntMap.intmap * Patheq.pathpt *
	     'a * Term.system * Patheq.problem -> 
	     ('a * Term.system * Patheq.problem) list) * 'a  ->
	    Term.init Term.IntMap.intmap * formula * database -> 
	    ('a proof * 'a more) option

	val setof : 
	    (Term.term * Term.init Term.IntMap.intmap * Patheq.pathpt *
	     'a * Term.system * Patheq.problem ->
	     ('a * Term.system * Patheq.problem) list) * 'a * 
	    formula * database ->
	    Patheq.pathpt * 
	    (Term.init Term.IntMap.intmap * formula ->
	     'a proof list)

	val followupMetavar :
	    (Term.term * Term.init Term.IntMap.intmap * Patheq.pathpt *
	     'a * Term.system * Patheq.problem -> 
	     ('a * Term.system * Patheq.problem) list) * 'a  ->
	    'a proof * formula * database -> ('a proof * 'a more) option

	val followupMetavarTr :
	    (Term.term * Term.init Term.IntMap.intmap * Patheq.pathpt *
	     'a * Term.system * Patheq.problem -> 
	     ('a * Term.system * Patheq.problem) list) * 'a  ->
	    Term.init Term.IntMap.intmap * 'a proof * 
	    formula * database -> ('a proof * 'a more) option

	val requeryFormula : 
	    (Term.term * Term.init Term.IntMap.intmap * Patheq.pathpt *
	     'a * Term.system * Patheq.problem ->
	     ('a * Term.system * Patheq.problem) list) * 'a  -> 
	     'a more * database -> ('a proof * 'a more) option

	val displayProof :
	    (((Patheq.pathpt option -> string) option -> 
	      (Term.init Term.IntMap.intmap option * Term.term) -> unit) * 
	     (Patheq.pathpt -> string) * 'a -> unit) * 
	    (string -> unit) ->  'a proof -> unit

	val readoutMetavar: 'a proof -> formula option
	val bindArg: formula * int * Patheq.pathpt -> Term.init Term.IntMap.intmap
    end
