(*
 * Matthew Stone
 * Dialup 0.1 - 3/96
 * modspec-sig.sml
 *
 * Signature file for collections of the modal inventories
 * described by the MODALITY signature.
 *
 * This module defines an abstract datatype for 
 * modalities, and functions to read and display
 * these entities.
 *
 * It also provides external interfaces to many
 * of the MODALITY functions.
 *)

signature MODSPEC =
    sig
	structure Modality : MODALITY
	structure Symbol : SYMBL

	(* a layer cake of modal inventories *)
	type collection

	(* abstraction of a box operator *)
	type absmod

	(* abstraction of a variable path *)
	type modState

	(* arg int gives the number of layers in the result *)
	val newCollection : unit -> collection

	(* return number of layers *)
	val levels: collection -> int

	(* return an abstraction of the top of the intth inventory *)
	val top: collection * int -> absmod

	(* return the intth inventory *)
	val inventory: collection * int -> Modality.inventory

	(*
	 * construct an abstract representation for
	 * the modality described in string list, using
	 * passed collection.  The string has the form
	 *           name {type} {level} deps*
	 * declaring name to be a modality with
	 * introspection rules given by type (default: K)
	 * at the layer given by level (default: 1)
	 * with inclusions [i] p -> name p for each
	 * modality [i] in deps*.
	 *)
	val readModality : (collection * string list -> 
			    absmod option * string list)

	(* show the modality's name *)
	val writeModality : absmod -> string
	val printModality : (string -> unit) -> absmod -> unit

	(* add the abstract modality to the collection *)
	val addModality : collection * absmod -> unit

	(* Modality.compile the different layers *)
	val compile : collection -> unit

	(* find the declared modality associated with string *)
	val lookup : collection * string -> absmod option 

	(* call Modality functions on the absmod *)
	val status : collection * absmod -> Modality.mclass
	val isInFunctional : collection * absmod -> bool
	val isFunctional : absmod -> bool

	(* describe what and where the absmod is *)
	val depth : absmod -> int
	val index : absmod -> int

	(* make a new empty abstract path *)
	val newState : collection -> modState

	(* 
	 * returns true unless extending path given by
	 * modState with a variable of type absmod
	 * will result in a path whose ambiguities
	 * the search engine can't cope with;
	 * when true, updates the modState accordingly.
	 *)
	val extendSafely : collection * modState * absmod -> bool
	val querySafely : collection * absmod * bool -> bool option
    end

