#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include "socket.h"

#define MAXCONN 10

SOCKET_SELECT *server;

int server_socket;
int connections[MAXCONN];
int stdin_open = 1;

void finish () 
{
   int i;

   close(server_socket);
   for (i = 0; i < MAXCONN; i++) 
      if (connections[i] != -1)
	 close (connections[i]);

   socket_select_free(server);
   exit(0);
}

void socket_read_callback(int fd)
{
   char buf[1000];
   int amount;
   int i;

   amount = read(fd, buf, 999);

   if (amount == 0) {
      printf("closing socket %d\n", fd);
      socket_select_remove(server, fd);
      close(fd);
      for (i = 0; i < MAXCONN; i++) {
	 if (connections[i] == fd)
	    while (i+1 < MAXCONN) {
	       connections[i] = connections[i+1];
	       i++;
	    }
      }
      if (i < MAXCONN) 
	 connections[i] = -1;

      if (connections[0] == -1 && !stdin_open)
	 finish ();
   }
   else if (amount < 0) {
      perror("something went wrong");
   }
   else {
      buf[amount] = 0;
      printf("%s", buf);
   }
}

void server_accept_callback(int fd)
{
   int newfd;
   int i;

   if ((newfd = socket_accept(fd)) == -1)
      exit(socket_errno);

   printf("connection established, connection %d\n", newfd);

   for (i=0; i<MAXCONN; i++)
      if (-1 == connections[i])
	 break;

   if (i == MAXCONN) {
      printf("Too many connections! Closing this one\n");
      close(newfd);
   }
   else {
      connections[i] = newfd;
      socket_select_add(server, newfd);
      socket_select_callback(server, newfd, SOCKET_CB_READ,
			     socket_read_callback);
   }
}

void input_callback(int fd)
{
   char buf[1000];
   int amount;
   int i;

   amount = read(fd, buf, 999);

   if (amount == 0) {
      printf("closing stdin (%d)\n", fd);
      socket_select_remove(server, fd);
      close(fd);
      stdin_open = 0;
      if (connections[0] == -1)
	 finish();
   }
   else if (amount < 0) {
      perror("something went wrong");
   }
   else {
      buf[amount] = 0;
      for(i = 0; i < MAXCONN; i++)
	 if (connections[i] != -1) {
	    write(connections[i], buf, amount);
	 }
   }
}

main(int argc, char *argv[])
{
   int port;
   int i;

   for (i=0; i < MAXCONN; i++)
      connections[i] = -1;

   if ((server_socket = socket_server_open(0)) == -1)
      exit(socket_errno);

   port = socket_port(server_socket);
   
   printf("Port is %d\n", port);

   server = socket_select_new();
   socket_select_add(server, server_socket);
   socket_select_callback(server, server_socket, SOCKET_CB_READ,
			  server_accept_callback);

   socket_select_add(server, 0); /* stdin */
   socket_select_callback(server, 0, SOCKET_CB_READ,
			  input_callback);

/*  socket_select_timeout(server, 180, 0); */

   (void) socket_select_loop(server);
   finish ();
}
