
/*
    socket.h
    Drew Olbrich, 1992
*/

#ifndef _SOCKET
#define _SOCKET

#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>

/* callbacks */

#define SOCKET_CB_READ          0
#define SOCKET_CB_WRITE         1
#define SOCKET_CB_EXCEPT        2

/* error codes */

#define SOCKET_ERR_SOCKET       1
#define SOCKET_ERR_BIND         2
#define SOCKET_ERR_LISTEN       3
#define SOCKET_ERR_HOST         4
#define SOCKET_ERR_CONNECT      5
#define SOCKET_ERR_ACCEPT       6
#define SOCKET_ERR_GETSOCKNAME  7
#define SOCKET_ERR_SETSOCKOPT   8
#define SOCKET_ERR_MALLOC       9
#define SOCKET_ERR_EXISTS       10
#define SOCKET_ERR_NOMATCH      11
#define SOCKET_ERR_SELECT       12
#define SOCKET_ERR_CLIENT       13
#define SOCKET_ERR_SERVER       14
#define SOCKET_ERR_PORT         15
#define SOCKET_ERR_OPEN         16
#define SOCKET_ERR_RENAME       17
#define SOCKET_ERR_READ         18
#define SOCKET_ERR_INVALIDCB    19

typedef struct {
    int fd;
    void (* callback_read)();
    void (* callback_write)();
    void (* callback_except)();
} SOCKET_SELECT_FD;

typedef struct { 
    int count;
    int max;
    SOCKET_SELECT_FD *sfd;
    fd_set fds_read;
    fd_set fds_write;
    fd_set fds_except;
    struct timeval timeout;
} SOCKET_SELECT;

extern int socket_errno;

int socket_server_open(int port);
int socket_client_open(char *host, int port);
int socket_accept(int fd);
int socket_port(int fd);
int socket_nodelay(int fd, int value);
SOCKET_SELECT *socket_select_new();
void socket_select_free(SOCKET_SELECT *s);
void socket_select_timeout(SOCKET_SELECT *s, int sec, int usec);
int socket_select_add(SOCKET_SELECT *s, int fd);
int socket_select_remove(SOCKET_SELECT *s, int fd);
int socket_select_callback(SOCKET_SELECT *s, int fd, int cb, void (*fn)());
int socket_select_loop(SOCKET_SELECT *s);
int socket_server_open_fn(char *fn);
int socket_client_open_fn(char *fn);
int socket_datagram_open(int port);

#endif

