
signature INTERPRETER =
    sig
	structure W : WIDGET
	structure S : SHELL
	structure D : DERIVE

	val newInterpreter : (W.root * string * D.Word.Formula.Patheq.Modspec.collection ->
			      bool CML.chan * unit CML.chan ->
			      unit)
    end

functor Interpreter (structure Fields : FIELDS;
		     structure Derive : DERIVE;
		     structure Corped : CORPED;
		     structure TNFE : TREE_N_FIELD_ED;
		     structure Morph : MORPHDB;
		     structure Lex : LEX;
		     structure Main : MAIN;
		     sharing TNFE.FE = Fields;
		     sharing Morph.Morph.Features = Derive.Itag.Tag.Features;
		     sharing Fields.FS = FieldStr = TNFE.TE.R.FS;
		     sharing Fields.S = TNFE.S = Corped.S = Shell;
		     sharing Fields.W = TNFE.W = Corped.W = Widget) : INTERPRETER =
    struct
	structure W = Widget
	structure S = Shell
	structure D = Derive

	fun deriveView (root, morphf, states) =
	    let
		val fields = D.stateOutFields
		val mkfields = #1 o (D.Itag.Tag.treeFields D.Itag.treeDataField)

		fun nodeView (name, (kind, data), channel, die) =
		    let
			val newchannel = CML.channel()
			fun prop () = (CML.send(channel,
						case (CML.accept newchannel) of
						    NONE => NONE
						  | SOME guy => SOME (kind, guy));
				       prop())
			    
			val result = 
			    Fields.mkFieldView(root,
					       Fields.FI{name = "node: " ^ name ^ "(" ^ kind ^ ")",
							 minwidth = SOME 400,
							 done = newchannel,
							 die = die,
							 action = NONE,
							 fields = (map Fields.cvtField
								   (ListUtil.zip (mkfields kind, data)))})
		    in
			CML.spawn prop;
			Shell.init result;
			result
		    end

		fun stateView (name, (tree, data), done, die) = 
		    let
			val result = 
			    TNFE.mkTreeNFieldView root
			    (TNFE.TFI{actions=TNFE.AL{init = ("words", [ "spud" ]),
						      edit = nodeView,
						      done = done,
						      die = die},
				      name = name,
				      minwidth = SOME 300,
				      fields = (map Fields.cvtField
						(ListUtil.zip (fields, data))),
				      tree = tree})
		    in
			Shell.init result;
			result
		    end

		fun mkinit (d) = 
		    let
			val (t, l) = D.writeState morphf d
		    in
			Corped.CE{key = hd l handle Hd => "<no sentence>",
				  data = (t, l)}
		    end

		val bogusinit =
		    (FieldStr.Tree{key = "bogus",
				   data = ("word", [ "spud" ]),
				   kids = nil},
		     [ "bogus", "bogus" ])

		val die = CML.channel()
		val done = CML.channel()

		val stepselect = 
		    Corped.mkRawCorped root (Corped.CI{actions=
						       Corped.AL{init = bogusinit,
								 group = fn _ => NONE,
								 write = fn (s,_) => s,
								 edit = stateView,
								 done = done,
								 die = die},
						       elinits = map mkinit states})

		val kill = CML.channel()
		val button =
		    Button.mkTextCmd root {rounded = false,
					   action = fn () => CML.send(kill, ()),
					   backgrnd = NONE,
					   foregrnd = NONE,
					   label = "Dismiss"}

		val layout = 
		    Box.mkLayout root (Box.VtCenter [Box.WBox stepselect,
						     Box.Glue {nat=5, min=5, max=SOME 5},
						     Box.WBox (Button.widgetOf button)])
		    
		val shell =
		    S.mkShell(Box.widgetOf layout, NONE, {win_name = SOME "generator output",
							  icon_name = SOME "output"})

		fun killfn () = case CML.accept kill of _ => (CML.send(die,false);
							      Shell.destroy shell; 
							      killfn())
	    in
		Shell.init shell;
		CML.spawn killfn;
		()
	    end
	
	local
	    fun mkILayout(root, collectionname, dataname, 
			  wordname, treename, morphname, entryname, 
			  entrywidget) =
		let
		    open Box Label
		    fun ll (s) = 
			WBox (widgetOf
			      (mkLabel root {label = s,
					     font = NONE,
					     foregrnd = NONE,
					     backgrnd = NONE,
					     align = W.HLeft}))

		    val l1 = ll("generator")
		    val l2 = ll("modalities: " ^ collectionname)
		    val l3 = ll("database: " ^ dataname)
		    val l4 = ll("words: " ^ wordname)
		    val l5 = ll("trees: " ^ treename)
		    val l6 = ll("morphs: " ^ morphname)
		    val l7 = ll("(entries " ^ entryname ^ ")")
		    val space = Glue{nat=5, min=5, max=SOME 5}
		in
		    Box.mkLayout root
		    (Box.VtLeft[space, l1, space, WBox entrywidget, space, l2, 
				space, l3, space, l4, 
				space, l5, space, l6, space, l7, space])
		end

	    fun mkEntryWidget(root, collection, 
			      database, words, trees, morphs, entries,
			      die, done) =
		let
		    open D FieldStr Corped
		    val fields = generandFields collection
		    val gen = generateHistory (words, trees, database) 
		    exception stop
		    
		    fun cvt (field as FieldStr.F{name,read}, str) = 
			case read (Lex.segment str) of
			    (SOME x, nil) => x
			  | _ => (print ("couldn't read field " ^ name ^ "\n");
				  raise stop)

		    fun action sl = 
			let
			    val g = assembleGenerand 
				(map cvt (ListUtil.zip (fields, sl)))
			in
			    case g of 
				SOME g' => 
				    deriveView (root, Morph.findForm morphs, gen g')
			      | NONE =>  (print "Aborted: Error in reading\n")
			end handle stop => (print "Aborted: Error in reading\n")
			
		    fun edit (name, data, channel, die) =
			let
			    val result =
				Fields.mkFields(root,
						Fields.FI{name = "generation instance",
							  minwidth =SOME 400,
							  done = channel,
							  die = die,
							  action = SOME action,
							  fields = (map Fields.cvtField
								    (ListUtil.zip (fields, data)))})
			in
			    Shell.init result;
			    result
			end

		    fun group s =
			(case FieldStr.segment (generandName, fields) s of
			     (FieldStr.GOT (name :: x), s') => (SOME (CE{key=name, data= name :: x}))
			   | (FieldStr.GOT nil, s') => (SOME (CE{key="bogus", data= nil}))
			   | (FieldStr.EMPTY, s') => (NONE)
			   | (FieldStr.ERROR e, _) =>
				 (print e; (NONE)))

		    fun write (x,s) = FieldStr.unsegment (generandName, fields) s
		in
		    mkSimpleCorped root (CI{actions = AL{init = gfieldInit,
							 edit = edit,
							 group = group,
							 write = write,
							 done = done,
							 die = die},
					    elinits = entries})
		end
	in
	    fun mkInterpreter(root, collectionname, collection)
		(dataname, database, wordname, words, 
		 treename, trees, morphname, morphs,
		 entryname, entries,
		 specpfx, die, done) =
		let
		    val temp = CML.channel ()
			
		    val (_, sset) = D.Word.Formula.setof(fn _ => nil, (), specpfx, database)

		    val entrywidget = 
			mkEntryWidget(root, collection, database, 
				      D.Word.specOrder sset words, 
				      D.Itag.specOrder sset trees,
				      morphs, entries, die, temp)
		    val layout =
			mkILayout(root, collectionname, dataname, wordname,
				  treename, morphname, entryname, entrywidget)

		    val result =
			Shell.mkShell(Box.widgetOf layout, 
				      NONE, {win_name = (SOME "generator"),
					     icon_name = (SOME "generator")})

		    fun moveOn () =
			(case CML.accept temp of _ =>
			     (Shell.destroy result;
			      CML.send(done, ());
			      moveOn ()))
		in
		    CML.spawn moveOn;
		    Shell.init result
		end
	end

	local
	    open Fields
	    fun check file = 
		let
		    open System.Unsafe.SysIO
		in
		    if access(file, [A_READ]) then
			true
		    else
			(print "Cannot access: ";
			 print file;
			 print "\n";
			 false)

		end

	    local
		val plist = ref ["egs/nixon.max", 
				 "egs/eng.tr",
				 "egs/eng.lex",
				 "egs/eng.mph",
				 "G $",
				 "egs/nixon.gen"]
	    in 
		fun params (collection) = 
		    [F{name = "database",
		       test = check,
		       init = nth(!plist, 0) },
		     F{name = "trees",
		       test = check,
		       init = nth(!plist, 1)},
		     F{name = "words",
		       test = check,
		       init = nth(!plist, 2)},
		     F{name = "morphology",
		       test = check,
		       init = nth(!plist, 3)},
		     F{name = "specificity context",
		       test = fn s => (case D.Word.Formula.readFormula 
					   collection (Lex.segment s) of 
					   (SOME _,nil) => true
					 | _ => false),
		       init = nth(!plist, 4)},
		     F{name = "tasks",
		       test = check,
		       init = nth(!plist, 5)}]

		fun updateDefaults newlist =
		    plist := newlist
	    end

	in
	    fun newInterpreter (root, collectionname, collection) =
		let
		    exception readError

		    fun readDatabase (file) =
			let
			    open D.Word.Formula
			    val rf = readFormula collection
			    fun reader (nil, db) = (db, false)
			      | reader (str, db) =
				(case rf str of
				     (SOME f, nil) =>
					 (addFormulas (db, [("anonymous", f)]), true)
				   | (SOME f, [ "." ]) =>
					 (addFormulas (db, [("anonymous", f)]), true)
				   | (SOME f, trail) =>
					 (print ("WARNING file " ^ file ^ 
						 ": unparsed trailing material:\n");
					  app (fn x => print (x ^ " ")) trail;
					  print "\n";
					  (addFormulas (db, [("anonymous", f)]), true))
				   | _ => (db, false))
			in
			    (case Main.readLoop (file, reader, makeDatabase(collection,nil)) of
				 NONE => 
				     (print "couldn't read database\n";
				      raise readError)
			       | SOME db => db)
			end
		    
		    fun readMorphs (file) =
			let
			    open Morph
			    val rm = Morph.readEntry
			    val db = newdb()
			    fun reader (nil, ()) = ((), false)
			      | reader (str, ()) =
				(case rm str of
				     (SOME m, _) =>
					 (addEntry db m, true)
				   | _ => ((), false))

			in
			    (case Main.readLoop (file, reader, ()) of
				 NONE =>
				     (print "couldn't read morphs\n";
				      raise readError)
			       | SOME () => 
				     db)
			end

		    fun readWords (file) = 
			let
			    open D.Word
			    val rw = readWord collection
			    fun reader (nil, lex) = (lex, false)
			      | reader (str, lex) =
				case rw str of
				    (SOME w, _) =>
					(addWord(lex, w), true)
				  | _ => (lex, false)
			in
			    (case Main.readLoop(file, reader, newLexicon()) of
				 NONE => 
				     (print "couldn't read lexicon\n";
				      raise readError)
			       | SOME lex => lex)
			end

		    fun readTrees (file) =
			let
			    open D.Itag
			    val rt = readEntry collection
			    fun reader (nil, trees) = (trees, false)
			      | reader (str, trees) =
				case rt str of
				    (SOME t, _) => (addEntry(trees, t), true)
				  | _ => (trees, false)
			in
			    (case Main.readLoop (file, reader, newTaglex()) of
				 NONE => 
				     (print "couldn't read trees\n";
				      raise readError)
			      | SOME trees => trees)
			end
		    
		    fun readTasks (file) =
			let
			    open D
			    val gf = generandFields collection
			    val rg = FieldStr.segment (generandName, gf)
			    fun reader (nil, gs) = (gs, false)
			      | reader (str, gs) =
				case rg str of
				    (FieldStr.GOT g, _) => 
					(Corped.CE{key=hd g handle Hd => "no name",
						   data = g} :: gs, 
					    true)
				  | _ => (gs, false)
			in
			    (case Main.readLoop (file, reader, nil) of
				 NONE =>
				     (print "couldn't read tasks\n";
				      raise readError)
			       | SOME gs => rev gs)
			end

		    fun readSpecpfx s =
			(case D.Word.Formula.readFormula collection (Lex.segment s) of
			     (SOME specpfx, nil) => specpfx
			   | (SOME specpfx, trail) =>
				 (print "WARNING: Unparsed trailing info in specificity test formula:\n";
				  app (fn x => (print (x ^ " "))) trail;
				  print "\n";
				  specpfx)
			   | _ => 
				 (print "ERROR: Couldn't parse specificity test formula \n";
				  raise readError))
			
		in
		    fn (die, done) =>
		    case getInfo(root, "interpreter parameters", 
				 params(collection), SOME 250) of
			SOME (l as [datafile, treefile, wordfile, morphfile, specpfx, taskfile]) =>
			    (updateDefaults l;
			     mkInterpreter(root, collectionname, collection)
			     (datafile, readDatabase datafile,
			      wordfile, readWords wordfile,
			      treefile, readTrees treefile,
			      morphfile, readMorphs morphfile,
			      taskfile, readTasks taskfile,
			      readSpecpfx specpfx, die, done)
			     handle readError =>
				 (CML.send(done, ())))
		      | _ => 
			    (print "Interpreter aborted\n";
			     CML.send(done, ()))
		end
	end
    end
