signature TOP = 
    sig 
	val i_main : string * string option -> OS.Process.status
	val main : string * string list -> OS.Process.status
    end

functor Toplevel (structure I:INTERPRETER;
		  structure Fields : FIELDS;
		  structure WE: WORD_EDIT;
		  structure TE: TREE_EDIT;
		  structure Main: MAIN;
		      
		  sharing I.D.Word.Formula.Patheq.Modspec = 
		      WE.Word.Formula.Patheq.Modspec =
		      TE.Tree.Formula.Patheq.Modspec) : TOP =
    struct
	exception badModspecs

	local
	    fun readModspecs (file) = 
		let
		    open TE.Tree.Formula.Patheq.Modspec
		    val new = newCollection ()
		    fun reader (str, ()) =
			(case readModality (new, str) of
			     (SOME m, _) =>
				 (addModality (new, m);
				  ((), true))
			   | _ => ((), false))
		in
		    (case Main.readLoop (file, reader, ()) of
			 NONE => raise badModspecs
		       | SOME () => (compile(new); new))
		end

	    datatype events = Trees | Words | Interp | Quit | Debug | Nodebug
		
	    fun mkSpudMenu (root, out) =
		let
		    open SimpleMenu MenuButton
		    val menu = 
			MENU [MenuItem ("edit trees", Trees),
			      MenuItem ("edit words", Words),
			      MenuItem ("new interpreter", Interp),
			      MenuItem ("quit", Quit),
			      MenuItem ("debug on", Debug), 
			      MenuItem ("debug off", Nodebug) ]

		    val (widget, event) =
			mkMenuButton root ("SPUD", menu)

		    fun loop () =
			(case CML.sync event of
			     x => (CML.send(out, x);
				   loop ()))
		in
		    CML.spawn loop;
		    widget
		end

	    fun setup (root, modspecname, modspecs, channel) =
		let
		    val newI = I.newInterpreter(root, modspecname, modspecs)
		    val newE = WE.newWordEdit(root, modspecs)
		    val newT = TE.newTreeEdit(root, modspecs)

		    fun process Quit =
			(Widget.delRoot root;
			 RunCML.shutdown (OS.Process.success))
		      | process Debug =
			(case Fields.getInfo(root, "set debug:", 
					     [Fields.F{name="flag", test = fn _ => true, init="derive"}],
					     SOME 200) of
			     SOME[flag] => Debug.debug flag
			   | _ => ())
		      | process Nodebug =
			(case Fields.getInfo(root, "unset debug:", 
					     [Fields.F{name="flag", test = fn _ => true, init="derive"}],
					     SOME 200) of
			     SOME[flag] => Debug.nodebug flag
			   | _ => ())
		      | process Interp = 
			let
			    val done = CML.channel()
			    val die = CML.channel ()
			    fun loop () =
				(case CML.recv done of _ => loop ())
			in
			    CML.spawn loop;
			    newI (die, done)
			end
		      | process Words = 
			let
			    val done = CML.channel()
			    val die = CML.channel ()
			    val shell = newE (die, done)
			    fun loop () =
				(case CML.recv done of _ => Shell.destroy shell;
				    loop ())
			in
			    CML.spawn loop;
			    ()
			end
		      | process Trees =
			let
			    val done = CML.channel()
			    val die = CML.channel ()
			    val shell = newT (die, done)
			    fun loop () =
				(case CML.recv done of _ => Shell.destroy shell;
				     loop ())
			in
			    CML.spawn loop;
			    ()
			end

		    fun pl () = pl(process(CML.recv channel);())
		in
		    CML.spawn pl
		end

	    fun tester (modspecname, modspecs) root =
		let
		    val out = CML.channel ()

		    val blunk = mkSpudMenu(root, out)
		    val place = Shell.mkShell(blunk, NONE,
					      {win_name = (SOME "spud"),
					       icon_name = (SOME "spud")})
		in
		    setup(root, modspecname, modspecs, out);
		    Shell.init place
		end

	in
	    fun i_main (file, addr) =
		(XDebug.init nil;
		 RunEXene.runWArgs 
		 (tester (file, readModspecs file))
		 {dpy= addr, timeq=SOME(20)};
		 OS.Process.success)

	    fun main (x,[y,z]) = i_main(y,SOME z)
	      | main (x,[y]) = i_main(y,NONE)
	      | main (x,_) = (print ("usage: " ^ x ^ " modfile  (display)\n");
				  OS.Process.failure)
	end
    end
