
signature TREE_EDIT =
    sig
	structure Tree : ITAG

	val newTreeEdit : (Widget.root * Tree.Formula.Patheq.Modspec.collection -> 
			   bool CML.chan * unit CML.chan ->
			   Shell.shell)
    end

functor TreeEdit (structure Tl : ITAG;
		  structure Fields : FIELDS;
		  structure TNFE : TREE_N_FIELD_ED;
		  structure Corped : CORPED;
		  sharing Fields = TNFE.FE) : TREE_EDIT =
    struct

	structure W = Widget
	structure S = Shell
	structure Tree = Tl

	open Widget 

	fun newTreeEdit (root, collection) =
	    let
		val fields = Tl.entryFields collection

		val mkfields = #1 o (Tl.Tag.treeFields Tl.treeDataField)

		fun nodeEdit (name, (kind, data), channel, die) =
		    let
			val newchannel = CML.channel ()
			fun prop () = (CML.send(channel,
						case (CML.recv newchannel) of 
						    NONE => NONE
						  | SOME guy => SOME(kind, guy));
				       prop())

			val result =
			    Fields.mkFields(root,
					    Fields.FI{name = "node: " ^ name ^ "(" ^ kind ^ ")",
						      minwidth = SOME 400,
						      done = newchannel,
						      die = die,
						      action = NONE,
						      fields = (ListPair.map Fields.cvtField 
								(mkfields kind, data))})

		    in
			CML.spawn prop;
			Shell.init result;
			result
		    end

		fun edit (name, (tree, data), channel, die) =
		    let
			val result = 
			    TNFE.mkTreeNField root
			    (TNFE.TFI{actions=TNFE.AL{init = ("words", [ "spud" ]),
						      edit = nodeEdit,
						      done = channel, 
						      die = die},
				      name = name,
				      minwidth = SOME 300,
				      fields = (ListPair.map Fields.cvtField
						(fields, data)),
				      tree = tree})
		    in
			Shell.init result;
			result
		    end

		val allFields = fields @ [ Tl.treeEntryField ]

		fun group s =
		    (case FieldStr.rawSegment ("entry", allFields) s of
			 (FieldStr.GOT list, s') =>
			     let
				 val data = List.take(list, length fields)
				 val treestuff = List.drop(list, length fields)
				 fun extractName (_ :: ":" :: s :: _) = s
				   | extractName (s :: _) = s
				   | extractName _ = "empty"
			     in
				 case Tl.Tag.segmentTree Tl.treeDataField (hd treestuff) of
				     (FieldStr.GOT tree, nil) => 
					 SOME(Corped.CE{key = extractName (hd data) handle Hd => "bogus",
							data = (tree, map FieldStr.merge data)})
				   | _ =>
					 NONE
			     end
		       | (FieldStr.ERROR e, _) => 
			     (print e;
			      NONE)
		       | _ => NONE)
			 
		fun write (key, (tree, data))  = 
		    let
			fun wd (FieldStr.Tree {key, data=(kind,values), kids}) =
			    Tl.Tag.writeForTree (Tl.treeDataField) (kind, values, map wd kids)
			val treestr = wd tree
		    in
			FieldStr.unsegment ("entry", allFields) (data @ [treestr])
		    end

	    in
		fn (die, done) => 
		Corped.mkCorped root (Corped.CI{actions = Corped.AL{init = Tl.fieldTreeInit,
								    edit = edit,
								    group = group,
								    write = write,
								    done = done,
								    die = die},
						elinits = nil})
	    end
    end

