
signature WORD_EDIT =
    sig
	structure Word : WORD

	val newWordEdit : (Widget.root * Word.Formula.Patheq.Modspec.collection -> 
			   bool CML.chan * unit CML.chan ->
			   Shell.shell)
    end

functor WordEdit (structure Word : WORD;
		  structure Fields : FIELDS;
		  structure Corped : CORPED) : WORD_EDIT =
    struct
	structure S = Shell
	structure W = Widget
	structure Word = Word

	open Widget Corped

	fun newWordEdit (root, collection) =
	    let
		val fields = Word.wordFields (collection);
		
		fun edit (name, data, channel, die) =
		    let
			val result =
			    Fields.mkFields(root,
					    Fields.FI{name = "word: " ^ name,
						      minwidth = SOME 400,
						      done = channel,
						      die = die,
						      action = NONE,
						      fields = (ListPair.map Fields.cvtField 
								(fields, data))})

		    in
			Shell.init result;
			result
		    end

		fun group s =
		    (case FieldStr.segment ("word", fields) s of
			 (FieldStr.GOT (name :: x), s') => (SOME (CE{key=name, data= name :: x}))
		       | (FieldStr.GOT nil, s') => (SOME (CE{key="bogus", data= nil}))
		       | (FieldStr.EMPTY, s') => (NONE)
		       | (FieldStr.ERROR e, _) =>
			     (print e; (NONE)))

		fun write (x,s) = FieldStr.unsegment ("word", fields) s
	    in
		fn (die, done) =>
		mkCorped root (CI{actions = AL{init = Word.fieldInit,
					edit = edit,
					group = group,
					write = write,
					done = done,
					die = die},
				  elinits = nil})
	    end
    end
