
functor Corped(structure MyFields : FIELDS) : CORPED =
    struct

	structure W = Widget
	structure S = Shell

	open CML Geometry Widget Shell Box Frame MenuButton 
	    SimpleMenu Shape Interact MyTextList Shell

	datatype 'a elt =
	    CE of {key: string,
		   data: 'a}

	datatype 'a actions =
	    AL of {init : 'a,
		   group : string list -> 'a elt option,
		   write : string * 'a -> string,
		   edit : string * 'a * 'a option CML.chan * bool CML.chan -> S.shell,
		   done : unit CML.chan,
		   die : bool CML.chan}

	datatype 'a corpinfo =
	    CI of {actions: 'a actions,
		   elinits: 'a elt list}

	datatype 'a action =
	    Edit of (int * 'a elt)
	  | Insert
	  | Save
	  | SaveAs
	  | Add
	  | Quit

	fun itemize (i as CE{key, data}) =  mkItem (key, i, Active false)

	fun setup (root, deal, listw, 
		   AL{init, group, write, edit, done, die}) =
	    let
		fun zombie () =
		    (case recv deal of
			 x => zombie())

		fun load (file) =
		    (let
			 val is = TextIO.openIn file
			 val isr = SIOA.wrapInstream is
			 fun next (old, ct) =
			     case SIOA.getclause isr of
				 "" => (old, ct) 
			       | s => 
				     case group (Lex1.segment s) of
					 NONE => next(old, ct)
				       | SOME more => next (more :: old, ct+1)

			 val (result, ct) = next(nil, 0)
		     in
			 print (Int.toString ct);
			 print " entries read.\n";
			 TextIO.closeIn is;
			 rev result
		     end		 
		 handle IO.Io {name,...} => (print "Error: ";
					     print name;
					     print "\n";
					     nil))

		fun save (file) =
		    (let 
			 val os = TextIO.openOut file
			 val osr = SIOA.wrapOutstream os
			 fun next (nil) = TextIO.closeOut os 
			   | next (CE{key,data} :: more) =
			     (SIOA.put(osr, write(key,data));
			      SIOA.put(osr, ".\n");
			      next more)
		     in
			 next (listItems listw);
			 TextIO.closeOut os
		     end	
		 handle IO.Io {name,...} => (print "Error: ";
					     print name;
					     print "\n"))


		fun process (Edit (code, CE{key, data}), (l, curfile)) =
		    let
			val editDone = channel()
			val editDie = channel()
			    
			val shell =  edit (key, data, editDone, editDie);

			fun loop () =
			    (case recv editDone of
				SOME result =>
				    (print "new value for ";
				     print (Int.toString code);
				     print "\n";
				     setData listw [ (code, CE{key=key, data=result}) ])
			      | NONE =>
				    ();
			     setActive listw [ (code, true) ];
			     setChosen listw [ (code, false) ];
			     destroy shell;
			     exit ())
		    in
			spawn loop;
			(editDie :: l, curfile)
		    end
		  | process (Insert, (l, curfile)) =
		    (print "insert!\n";
		     case MyFields.getFile(root, "file to insert", true) of
			 SOME file =>
			     (addItems listw (List.map itemize (load file)))
		       | _ => ();
		     (l, curfile))
		  | process (SaveAs, (l, curfile)) =
		    (print "save as!\n";
		     case MyFields.getFile(root, "save as", false) of
			 SOME file =>
			     (save file;
			      (l, file))
		       | _ => (l, curfile))
		  | process (Save, (l, curfile)) =
		    (print "save!\n";
		     save curfile;
		     (l, curfile))
		  | process (Add, (l, curfile)) =
		    (print "add!\n";
		     case MyFields.getInfo (root, "new guy", [ MyFields.F{name = "name",
								      init = "",
								      test = fn _ => true} ],
					    SOME 250) of
			 SOME [ key ] => 
			     addItems listw [ itemize (CE{key=key, data=init}) ]
		       | _ => ();
		      (l, curfile))
		  | process (Quit, (l, curfile)) =
		    (app (fn x => send (x,false)) l;
		     send(done, ());
		     TraceCML.unwatch(getTid());
		     zombie())

		fun loop (lc) =
		    loop (process (select[recvEvt deal,
					  wrap (recvEvt die, fn _ => Quit)],
				   lc))
	    in
		TraceCML.watch ("editor", spawn (fn () => loop (nil, "output")))
	    end

	fun mkMyTextList (root, deal, eltinits) =
	    let
		val (_, list_items) = 
		    foldl (fn (elt, (i, items)) => (i+1, (itemize elt) :: items))
		    (0, nil)
		    eltinits

		val textList =
		    mkVList root {mode = MultiSet,
				  backgrnd = NONE,
				  foregrnd = NONE,
				  items = rev list_items}

		fun loop () =
		    (case sync (evtOf textList) of
			 (Set (code, element)) =>
			     (send(deal, Edit (code, element));
			      setActive textList [ (code, false) ];
			      loop())
		       | (Unset _) => 
			     (loop()))
	    in
		spawn loop;
		textList
	    end

	fun mkCorpedMenu (root, deal) =
	    let
		val menu = 
		    MENU [MenuItem ("insert", Insert),
			  MenuItem ("save", Save),
			  MenuItem ("save as", SaveAs),
			  MenuItem ("add", Add),
			  MenuItem ("quit", Quit)]

		val (widget, event) =
		    mkMenuButton root ("stuff", menu)

		fun loop () =
		    (case sync event of
			 x => (send(deal, x);
			       loop ()))
	    in
		spawn loop;
		widget
	    end

	fun mkCorpedLayout (root, textlist, menu) =
	    let
		val glue = Glue{nat=5, min=1, max=NONE}
		val space = Glue{nat=5, min=5, max=SOME 5}
		val frame = 
		    mkFrame 
		    {color = SOME (EXB.blackOfScr (screenOf root)),
		     width = 2,
		     widget = MyTextList.widgetOf textlist}
	    in
		mkLayout root
		(VtCenter [space,
			   HzCenter[WBox (mkRigid menu),
				    glue],
			   space,
			   HzCenter[space,
				    WBox (Frame.widgetOf frame),
				    space],
			   space])
	    end

	local 
	    fun mkEitherCorped menup root (CI{actions,elinits}) =
		let
		    val deal = channel ()
		    val tlw = mkMyTextList (root, deal, elinits)
		    
		    val returnval =
			if menup then
			    Box.widgetOf
			    (mkCorpedLayout(root, tlw, mkCorpedMenu (root, deal)))
			else
			    Frame.widgetOf
			    (mkFrame 
			     {color = SOME (EXB.blackOfScr (screenOf root)),
			      width = 2,
			      widget = MyTextList.widgetOf tlw})
		in
		    setup(root,deal,tlw,actions);
		    returnval
		end
	in
	    val mkRawCorped = fn u => mkEitherCorped false u
	    val mkSimpleCorped = fn u => mkEitherCorped true u
	end

	fun mkCorped root (stuff) =
	    let
		val corped = mkSimpleCorped root stuff
		val shell = mkShell (corped, NONE, 
				     {win_name = SOME "spud editor",
				      icon_name = SOME "spud editor"})
	    in
		init shell;
		shell
	    end
    end
