structure Fields : FIELDS =
    struct
	
	structure FS = FieldStr

	open CML Geometry Widget Shell Box Frame Label StrEdit Shape Interact Button

	local 
	    open Interact
	in
	    fun cleanse (KEY_Press (NoSymbol, state)) = 
		(print "Cleanse: got bogus press!\n"; NONE)
	      | cleanse (KEY_Release (NoSymbol, state)) = 
		(print "Cleanse: got bogus release!\n"; NONE)
	      | cleanse x = SOME x
	end

	datatype field = 
	    F of {name: string,
		  init: string,
		  test: string -> bool}

	datatype fieldinfo =
	    FI of {name: string,
		   minwidth: int option,
		   done: string list option chan,
		   die: bool CML.chan,
		   action: (string list -> unit) option,
		   fields: field list}

	datatype fieldmsg =
	    New of ((string -> bool) * str_edit * label * int * 
		    kbd_msg addr_msg event * kbd_msg addr_msg chan)
	  | Input of kbd_msg addr_msg
	  | Mouse of int
	  | OK
	  | Go
	  | Cancel

	datatype fieldsetup =
	    FS of {label : widget,
		   frame : widget}

	datatype fielder =
	    FR of {editor : str_edit,
		   label : label,
		   test : string -> bool,
		   num : int,
		   send : kbd_msg addr_msg chan}

	datatype fieldbox =
	    FB of {bef: fielder list,
		   notbef: fielder list,
		   active : int,
		   activechan: kbd_msg addr_msg chan}

	fun check (FR{editor, test, ...}) =  
	    let
		val s = getString editor
	    in 
		(test s, s)
	    end 

	fun addbef (nil, l, sofar) = rev (l :: sofar)
	  | addbef ((f as FR{num,...})::more, g as FR{num = num2,...}, sofar) =
	    if num > num2 then
		addbef(more, g, f :: sofar)
	    else
		U.rappend (g :: sofar) (f :: more) 

	fun addafter (nil, l, sofar) = rev (l :: sofar)
	  | addafter ((f as FR{num,...})::more, g as FR{num = num2,...}, sofar) =
	    if num < num2 then
		addafter(more, g, f :: sofar)
	    else
		U.rappend (g :: sofar) (f :: more)

	fun add(bef, notbef, active, g as FR{num,...}) =
	    if active < num then
		(bef, addafter(notbef, g, nil))
	    else
		(addbef(bef, g, nil), notbef)

	fun grey (root) = 
	    (EXB.colorOfScr
	     (screenOf root)
	     (EXB.CMS_Name "grey40"))

	fun black (root) = 
	    (EXB.colorOfScr
	     (screenOf root)
	     (EXB.CMS_Name "black"))

	fun drawActive (old, new, root) =
	    (Label.setForeground old (grey root);
	     Label.setForeground new (black root))

	fun advance (bef, l as [ FR{label=lo,...} ], root) =
	    (case U.rappend bef l of 
		 nil => NONE
	       | l as (FR{editor, label, num, send,...} :: _) => 
		     (drawActive (lo, label, root);
		      SOME (FB{bef = nil,
			       notbef = l,
			       active = num,
			       activechan = send})))
	  | advance (bef, (e as FR{label=lo,...}) :: 
		     (l as (FR{editor, label, num, send,...} :: _)),
		     root) =
	    (drawActive (lo, label, root);
	     SOME (FB{bef = e :: bef,
		      notbef = l,
		      active = num,
		      activechan = send}))
	  | advance _ = NONE	    

	fun popto(bef, notbef as (FR{label=old,...} :: _), root, i) =
	    let
		fun popmore(b, a as ((e1 as FR{label=w1, send=c1, num=n1,...}) :: m1)) =
		    if i = n1 then
			(drawActive (old, w1, root);
			 SOME (FB{bef = b,
				  notbef = a,
				  active = i,
				  activechan = c1}))
		    else 
			if i > n1 then
			    popmore(e1 :: b, m1)
			else 
			    (case b of 
				 nil => NONE
			       | (e2 as FR{num=n2,...}) :: more => 
				     if i > n2 then
					 NONE
				     else popmore(more, e2 :: a))

		  | popmore(notbef, nil) = NONE
	    in
		popmore(bef, notbef)
	    end
	  | popto _ = NONE

	fun setup (fieldmsgs, root, done, action) =
	    let
		val empty = channel () 

		fun discard (event) = 
		    let fun loop () = 
			(sync event; print "empty!\n"; loop ())
		    in loop
		    end
				   
		val init = FB{bef = nil,
			      notbef = nil,
			      active = ~1,
			      activechan = empty }

		fun zombie () =
		    (case recv fieldmsgs of
			 x => zombie())

		fun dieNow (result) =
		    (send(done, result);
		     TraceCML.unwatch(getTid());
		     zombie ())

		fun withans f (bef,notbef) =
		    let
			val (good, answer) = 
			    foldr  (fn (fr,(g,a)) => 
				   let val (g', e) = check fr
				   in (g' andalso g, e :: a)
				   end) 
			    (true, nil)
			    (U.rappend bef notbef)
		    in
			if good then
			    (f (SOME answer))
			else
			    (print "Error in data!\n")
		    end

		val finish = withans dieNow

		fun process (f as FB{bef, notbef, active, activechan}) =
		    (case recv fieldmsgs of
			 New (test, guy, label, i, oldevents, newevents) =>
			     let
				 val (b', n') = add(bef, notbef, active, 
						    FR{test = test,
						       editor = guy, 
						       label = label,
						       num = i,
						       send = newevents})
				 val (na, nc) = 
				     if active = ~1 then
					 (Label.setForeground label 
					  (black (rootOf (Label.widgetOf label)));
					  (i, newevents))
				     else
					 (active, activechan)
			     in
				 process(FB{bef = b',
					    notbef = n',
					    active = na,
					    activechan = nc})
			     end

		       | Input stuff => 
			     (case cleanse(msgBodyOf stuff) of
				  SOME (KEY_Press info) =>
				      (case (lookupString defaultTranslation info 
					     handle _ => "") of
					   "\t" =>
					       (case advance(bef, notbef, root) of
						    NONE => process f
						  | SOME new => process new)
					 | "\n" => 
					       (finish (bef, notbef);
						process f)
					 | "\^M" => 
					       (finish (bef, notbef);
						process f)
					 | "\^C" => 
					       dieNow (NONE)
					 | x => 
					       (send (activechan, stuff);
						process f))
				| SOME (other) => 
				      (send (activechan, stuff);
				       process f)
				| NONE => (process f))
		       | Mouse i => 
			     (case popto(bef, notbef, root, i) of
				  NONE => process f
				| SOME new => process new)
		       | Go => 
			     (case action of 
				  NONE => ()
				| SOME a => 
				      withans (fn NONE => () | SOME x => a x)
				      (bef, notbef);
			      process f)
		       | OK => 
			     (finish (bef, notbef);
			      process f)
		       | Cancel => 
			     dieNow (NONE))
	    in
		spawn (discard (recvEvt empty));
		TraceCML.watch ("process", spawn (fn () => process init))
	    end		
		
	fun makeFieldEditor (root, fieldmsgs) (F{name, test, init}, i) =
	    let
		val blackc = EXB.blackOfScr (screenOf root)
		val label =
		    mkLabel root {label = name,
				  foregrnd = SOME (grey root),
				  backgrnd = NONE,
				  font = NONE,
				  align = HRight}
		val editor =
		    mkStrEdit root {initval = init,
				    foregrnd = NONE,
				    backgrnd = NONE,
				    minlen = 0}

		val (wrapper1, keyevent) =
		    filterKey (StrEdit.widgetOf editor)
		    
		val (wrapper2, mousevent) =
		    filterMouse wrapper1

		fun interpose (e,c) =
		    let 
			fun loop () =
			    case sync e of
				x => (send(c,x);
				      (case msgBodyOf x of
					   MOUSE_Down _ => send(fieldmsgs, Mouse i)
					 | MOUSE_FirstDown _ => send(fieldmsgs, Mouse i)
					 | _ => ());
				      loop ())
		    in
			loop
		    end

		val frame =
		    mkFrame {widget = wrapper2,
			     color = SOME (grey root),
			     width = 2}
	    in
		spawn(fn () => let val (e,c) = (sync keyevent)
			       in  send(fieldmsgs, New(test, editor, label, i, e, c));
				   exit() 
			       end);
		spawn(fn () => let val j = (sync mousevent) 
			       in  spawn (interpose j);
				   exit()
			       end);
		(FS{label = Label.widgetOf label,
		    frame = Frame.widgetOf frame},
		 i + 1)
	    end

	fun mkFieldLayout (root, name, fieldmsgs, fielders, 
			   action, minwidth) =
	    let
		val vglue = Glue {nat=5, min=1, max=NONE}

		val namelabel =
		    mkLabel root {label = name,
				  font = NONE,
				  foregrnd = SOME (black root),
				  backgrnd = NONE,
				  align = HLeft}

		val ok =
		    mkTextCmd root {rounded = false,
				    action = fn () => send(fieldmsgs, OK),
				    backgrnd = NONE,
				    foregrnd = NONE,
				    label = "OK"}

		val cancel =
		    mkTextCmd root {rounded = false,
				    action = fn () => send(fieldmsgs, Cancel),
				    backgrnd = NONE,
				    foregrnd = NONE,
				    label = "NO"}

		fun makelines (FS{label, frame,...}, (labels, values)) =
		    (vglue :: (WBox (mkRigid label)) :: labels,
		     vglue :: (WBox frame) :: values)

		val (labels, values) = 
			 foldr makelines ([vglue], [vglue]) fielders

		val bottomRow = 
		    (case action of 
			 NONE => 
			     (HzCenter [Glue {nat=5, min=5, max=SOME 5},
					WBox (Button.widgetOf cancel),
					vglue,
					WBox (Button.widgetOf ok),
					Glue {nat=5, min=5, max=SOME 5}])
		       | SOME _ => 
			     (HzCenter [Glue {nat=5, min=5, max=SOME 5},
					WBox (Button.widgetOf cancel),
					vglue,
					WBox (Button.widgetOf 
					      (mkTextCmd root {rounded=false,
							       action=fn () => send(fieldmsgs, Go),
							       backgrnd = NONE,
							       foregrnd = NONE,
							       label = "GO"})),
					vglue,
					WBox (Button.widgetOf ok),
					Glue {nat=5, min=5, max=SOME 5}]))
			 
	    in
		mkLayout root
		    (VtCenter[vglue,
			      WBox (Label.widgetOf namelabel),
			      HzCenter [case minwidth of 
					    NONE => vglue 
					  | SOME w => Glue{nat=w, min=w, max=NONE}],
			      vglue,
			      HzCenter [Glue {nat=5, min=5, max=SOME 5},
					VtRight labels,
					Glue {nat=5, min=5, max=SOME 5},
					VtCenter values,
					Glue {nat=5, min=5, max=SOME 5}],
			      vglue,
			      bottomRow,
			      vglue
			      ])
	    end

	fun mkSimpleFields (root, FI{name,minwidth,done,die,fields,action}) =
	    let
		val fieldmsgs = channel ()
		val (fielders,_) = 
		    U.foldstate (makeFieldEditor (root, fieldmsgs)) (fields, 0)
		val mfe = mkFieldLayout (root, name, fieldmsgs, fielders, 
					 action, minwidth)

		fun sendalong event = 
		    let fun loop () =
			let val e = (sync event)
			in  send (fieldmsgs, Input e);
			    loop ()
			end
		    in loop ()
		    end

		val (wrapper, event) = 
		    filterKey (Box.widgetOf mfe)

	    in
		spawn(fn () => (case recv die of
				    true => send(fieldmsgs, OK)
				  | false => send(fieldmsgs, Cancel)));
		TraceCML.watch ("send", spawn(fn () => sendalong (#1 (sync event))));
		setup(fieldmsgs,root,done,action);
		wrapper
	    end

	fun mkFields (root, fi as FI{name,...}) =
	    let
		val wrapper = mkSimpleFields (root, fi)
		val shell = mkShell
		    (wrapper, NONE, {win_name = SOME name, 
				     icon_name = SOME name})
	    in
		shell
	    end

	fun getInfo (root, name, info, minwidth) =
	    let
		val done = CML.channel ()
		val die = CML.channel ()
		val guy = mkFields(root, FI{name = name,
					    minwidth = minwidth,
					    done = done,
					    die = die,
					    fields = info,
					    action = NONE});
	    in
		Shell.init guy;
		CML.recv done before
		Shell.destroy guy
	    end

	fun getFile (root, name, toRead) =
	    let
		fun check file =
		    let
			open OS.FileSys
		    in
			if (not toRead) orelse access(file, [ A_READ ]) then
			    true
			else
			    (print "Cannot access: ";
			     print file;
			     print "\n";
			     false)
		    end

		val fields = [ F {name = "file",
				  test = check,
				  init = "default"} ]
	    in
		case getInfo(root, name, fields, SOME 250) of
		    SOME [ file ] => SOME file
		  | _ => NONE
	    end

	fun cvtField (f as FS.F{name,...}, init) =
	    F{name = name,
	      init = init,
	      test = FS.test f}

	fun mkSimpleFieldView (root, FI{name, done, die, fields,...}) =
	    let
		val space = Glue {nat=5, min = 5, max=SOME 5}
		fun ll (F{name, init, ...}, guys) =
		    (WBox (Label.widgetOf 
			   (Label.mkLabel root {label = name ^ ": " ^ init,
						font = NONE,
						foregrnd = NONE,
						backgrnd = NONE,
						align = W.HLeft}))) ::
		    space :: guys

		val dismiss =
		    Button.mkTextCmd root {rounded = false,
					   action = fn () => send(done, NONE),
					   backgrnd = NONE,
					   foregrnd = NONE,
					   label = "Dismiss"}

		fun wait () =
		    (case recv die of _ => (send(done, NONE); wait()))
	    in
		spawn wait;
		Box.widgetOf
		(Box.mkLayout root 
		 (VtLeft (space :: (foldr ll [WBox (Button.widgetOf dismiss), space] fields))))
	    end

	fun mkFieldView (root, fi as FI{name,...}) =
	    let
		val wrapper = mkSimpleFieldView (root, fi)
		val shell = mkShell
		    (wrapper, NONE, {win_name = SOME name, 
				     icon_name = SOME name})
	    in
		shell
	    end

    end

