(* text_list.sml
 *
 * COPYRIGHT (c) 1991,1992 by AT&T Bell Laboratories
 *
 * List widget, for text lists.
 *
 * modified ms 6/97
 *)

signature MY_TEXT_LIST =
  sig

    structure W : WIDGET

    exception BadIndex
    exception MultipleChoices

    type 'a text_list
    datatype 'a list_evt = 
	Set of int * 'a 
      | Unset of int * 'a

    (* type 'a list_item = (string * 'a * W.wstate) 
     *)
    type 'a list_item
    val mkItem : (string * 'a * W.wstate) -> 'a list_item

    datatype list_mode = OneSet | MultiSet

    val mkHList : W.root -> {
	    mode : list_mode,
	    backgrnd : W.EXB.color option,
	    foregrnd : W.EXB.color option,
	    items : '2a list_item list
	  } -> '2a text_list
    val mkVList : W.root -> {
	    mode : list_mode,
	    backgrnd : W.EXB.color option,
	    foregrnd : W.EXB.color option,
	    items : '2a list_item list
	  } -> '2a text_list

    val evtOf : 'a text_list -> 'a list_evt CML.event
    val widgetOf : 'a text_list -> W.widget

    val setChosen : 'a text_list -> (int * bool) list -> unit
    val setActive : 'a text_list -> (int * bool) list -> unit
    val setData : 'a text_list -> (int * 'a) list -> unit

    val getChosen : 'a text_list -> int list
    val getState : 'a text_list -> W.wstate list

    val addItems : 'a text_list -> 'a list_item list -> unit
    val listItems : '1a text_list -> '1a list
  end (* MY_TEXT_LIST *)

structure MyTextList : MY_TEXT_LIST = 
  struct

    structure W = Widget

    open Geometry WidgetSet

    type 'a list_item = (string * 'a * W.wstate)
    fun mkItem x = x
    datatype 'a list_evt = 
	Set of int * 'a 
      | Unset of int * 'a

    datatype 'a item_msg =
	Pick of bool
      | IChange of 'a
      | Report of 'a list * 'a item_msg CML.chan list * 'a list CML.chan

    datatype 'a list_msg =
	Add of 'a list_item list
      | Change of (int * 'a) list
      | ListAll of 'a list CML.chan

    datatype list_mode = OneSet | MultiSet
    datatype 'a text_list = 
	TL of {
	       ch : 'a list_msg CML.chan,
	       widget : W.widget,
	       wset : widget_set,
	       evt : 'a list_evt CML.event
	       }

    fun dummy _ = ()
    fun addListItem (root,backgrnd,foregrnd,ch) ((label, v, state), (widgets, channels, ct)) = 
	let
	    open Toggle
	    val w = mkToggleText root {
				       state=state,
				       rounded=false,
				       action = dummy,
				       backgrnd = backgrnd,
				       foregrnd = foregrnd,
				       label = label
				       }
	    val itemChan = CML.channel ()

	    fun loop v =
		(case CML.recv itemChan of
		     Pick true => (CML.send(ch, Set(ct,v));
				   loop v)
		   | Pick false => (CML.send(ch, Unset(ct,v));
				    loop v)
		   | IChange new => loop new
		   | Report (sofar, nil, out) => 
			 (CML.send(out, (rev sofar));
			  loop v)
		   | Report (sofar, next :: rest, out) => 
			 (CML.send(next, Report(v :: sofar, rest, out));
			  loop v))

	    fun pick_fn b = (CML.send (itemChan, Pick b);
			     setState(w,b))
	    fun active_fn b = setActive (w,b)
	in 
	    CML.spawn (fn () => loop v);
            ({
              widget=W.ignoreMouse(widgetOf w),
              state=state,
              pick_fn=pick_fn,
              active_fn=active_fn
	      } :: widgets,
	     itemChan :: channels,
	     ct + 1)
	end

    fun addDummy (root,backgrnd,foregrnd,ch) =
	let
	    open Toggle
	    val w = mkToggleText root {
				       state= W.Inactive false,
				       rounded=false,
				       action = dummy,
				       backgrnd = backgrnd,
				       foregrnd = foregrnd,
				       label = ""
				       }

	    val itemChan = CML.channel ()

	    fun loop () =
		(case CML.recv itemChan of
		     Report (sofar, nil, out) => 
			 (CML.send(out, (rev sofar)); loop ())
		   | Report (sofar, next :: rest, out) => 
			 (CML.send(next, Report(sofar, rest, out)); loop ())
		   | _ => loop ())

	    fun pick_fn b = setState(w,b)
	    fun active_fn b = setActive (w,b)
	in
	    CML.spawn loop;
            ([{
              widget=W.ignoreMouse(widgetOf w),
              state= W.Inactive false,
              pick_fn=pick_fn,
              active_fn=active_fn
	      }],
	     [itemChan],
	     1)
	end

    fun maxSize (({widget,...} : WidgetSet.set_item), SIZE{wid,ht}) = let
          val SIZE{wid=w,ht=h} = Widget.natSize widget
          in SIZE{wid=Int.max(wid,w),ht=Int.max(ht,h)} end

    fun setSize sz ({widget,state,pick_fn,active_fn}) = 
	({
          widget=Shape.freeSize (widget,sz), 
          state=state,
          pick_fn=pick_fn,
          active_fn=active_fn
        })

    fun mkList layout root {mode, backgrnd, foregrnd, items} = 
	let
	    val ch = CML.channel()
	    val pvt = CML.channel()

	    val addItem = addListItem (root, backgrnd, foregrnd, ch)

	    val (widgets, channels, count) = 
		(case (foldl addItem (nil, nil, 0) items) of
		     (nil, nil, 0) => addDummy (root, backgrnd, foregrnd, ch)
		   | s => s)
		
	    val sz = foldl maxSize (SIZE{wid=0,ht=0}) widgets
	    val items' = map (setSize sz) (rev widgets)
	    val (wset,wl) = case mode of
		OneSet => WidgetSet.mkSingleSet root items'
	      | MultiSet => WidgetSet.mkMultiSet root items'
	    val box = Box.mkLayout root (layout(map Box.WBox wl))

	    fun process (channels, count) =
		(case CML.recv pvt of
		     Add items =>
			 let
			     val (newguys, newchannels, newcount) =
				 (foldl addItem (nil, channels, count) items)
			     fun mapb (c,l) = if c < count then Box.mapBox box l
					      else mapb (c-1, c::l)
			 in
			     Box.insert box
			     (count,
			      map Box.WBox
			      (WidgetSet.insert wset 
			       (count, map (setSize sz) (rev newguys))));
			     mapb (newcount-1,nil);
			     process (newchannels, newcount)
			 end
		   | Change guys =>
			 let
			     fun change (i,v) =
				 (CML.send (List.nth (rev channels, i), IChange v)
				  handle Nth => raise BadIndex)
			 in
			     app change guys;
			     process (channels, count)
			 end
		   | ListAll out => 
			 (case channels of 
			      nil => CML.send(out, nil)
			    | head :: rest => CML.send(head, Report(nil, rest,out));
			  process (channels, count)))
	in
	      CML.spawn (fn () => process (channels, count));
	      TL {
		  ch = pvt,
		  widget = Box.widgetOf box,
		  wset = wset,
		  evt = CML.recvEvt ch
		  }
	end

    val mkHList = fn u => mkList Box.HzCenter u
    val mkVList = fn u => mkList Box.VtCenter u

    fun evtOf (TL{evt,...}) = evt
    fun widgetOf (TL{widget,...}) = widget

    fun setChosen (TL{wset,...}) arg = WidgetSet.setChosen wset arg
    fun setActive (TL{wset,...}) arg = WidgetSet.setActive wset arg
    fun setData (TL{ch,...}) arg = CML.send(ch, Change arg)

    fun getChosen (TL{wset,...}) = WidgetSet.getChosen wset
    fun getState (TL{wset,...}) = WidgetSet.getState wset

    fun addItems (TL{ch,...}) arg = CML.send(ch, Add arg)
    fun listItems (TL{ch,...})  = 
	let
	    val out = CML.channel ()
	in 
	    CML.send(ch, ListAll out);
	    CML.recv out
	end

end (* TextList *)
