(*
 *  rec list: modified from (modified) text_list ms 6/97
 *)

structure RecList : REC_LIST =
    struct
	structure W = Widget
	structure FS = FieldStr

	open Widget Geometry WidgetSet

	datatype 'a list_evt = 
	    Set of int * string * 'a 
	  | Unset of int * string * 'a

	datatype 'a item_msg =
	    Pick of bool
	  | IChange of 'a
	  | Report of 'a FS.node list * 'a item_msg CML.chan list * 'a item_msg CML.chan
	  | Return of 'a FS.node list

	datatype 'a gnode =
	    TG of {key: string,
		   data: 'a,
		   kidcs: 'a item_msg CML.chan list}

	datatype 'a rec_event =
	    Add of (int * 'a FS.node)
	  | Change of (int * 'a) list
	  | ListAll of 'a item_msg CML.chan

	datatype 'a rec_list = 
	    RL of {
		   ch : 'a rec_event CML.chan,
		   widget : W.widget,
		   wset : widget_set,
		   evt : 'a list_evt CML.event
		   }

	fun dummy _ = ()

	datatype list_mode = OneSet | MultiSet

	fun mkGeom (add, root, label, channel, ct, nil) =
	    let
		val w = W.ignoreMouse(Toggle.widgetOf label);

		fun pick_fn b = (CML.send (channel, Pick b);
				 Toggle.setState(label,b))
		fun active_fn b = Toggle.setActive (label,b)
	    in
		map Box.WBox (add (ct, [WidgetSet.mkSetItem
					{widget = w, state = Active false, 
					 pick_fn = pick_fn, active_fn = active_fn}]))
	    end
	  | mkGeom (add, root, label, channel, ct, kids) =
	    let
		val wi = W.ignoreMouse(Toggle.widgetOf label);
		fun pick_fn b = (CML.send (channel, Pick b);
				 Toggle.setState(label,b))
		fun active_fn b = Toggle.setActive (label,b)

		val wm = Box.mkLayout root
		    (Box.VtCenter ((map Box.WBox (add (ct, [WidgetSet.mkSetItem 
							   {widget = wi, state = Active false,
							    pick_fn = pick_fn, active_fn = active_fn}])))
				   @ [Box.HzTop kids]))
		
	    in
		[ Box.WBox (Box.widgetOf wm) ]
	    end

	fun addNodeView (add,root,backgrnd,foregrnd,ch) 
	    (FS.Tree{key, data:'2a, kids}, 
	     (layout, allchannels, topchannels, ct)) =
	    let
		open Toggle
		val w = mkToggleText root 
		    {state = Active false,
		     rounded = false,
		     action = dummy,
		     backgrnd = backgrnd,
		     foregrnd = foregrnd,
		     label = key}
		val itemChan = CML.channel () : '2a item_msg CML.chan

		val (sublayout, morechannels, subchannels, subct) =
		    foldl (addNodeView (add,root,backgrnd,foregrnd,ch)) 
		    (nil, itemChan :: allchannels, nil, ct) kids

		fun get (nil, data) = FS.Tree{key=key, data=data, kids=nil}
		  | get (first :: more, data) = 
		    (CML.send(first, Report(nil, more, itemChan));
		     case CML.recv itemChan of
			 Return kids => FS.Tree{key=key, data=data, kids=kids}
		       | _ => (print "kids lost!\n";
			       FS.Tree{key=key, data=data, kids=nil}))

		fun loop (v as TG{key,data,kidcs})  =
		    (case CML.recv itemChan of
			 Pick true => (CML.send (ch, Set(subct, key, data));
				       loop v)
		       | Pick false => (CML.send (ch, Unset(subct, key, data));
					loop v)
		       | IChange new => (loop (TG{key=key, data=new, kidcs=kidcs}))
		       | Report (sofar, nil, out) =>
			     (CML.send (out, Return (get (kidcs, data) :: sofar));
			      loop v)
		       | Report (sofar, next :: rest, out) => 
			     (CML.send(next, Report((get (kidcs, data)) :: sofar, 
						    rest, out));
			      loop v)
		       | Return l => loop v)

		val box = mkGeom(add, root, w, itemChan, subct, rev sublayout)
	    in
		CML.spawn (fn () => loop (TG{key=key, data=data, kidcs=subchannels}));
		(box @ layout, 
		 morechannels,
		 itemChan :: topchannels, subct+1)
	    end

	fun mkRList root {mode, backgrnd, foregrnd, item} =
	    let
		val ch = CML.channel ()
		val pvt = CML.channel ()

		val (wset, _) = case mode of
		    OneSet => WidgetSet.mkSingleSet root nil
		  | MultiSet => WidgetSet.mkMultiSet root nil

		val (outbox, allchannels, topchannels, count) =
		    (addNodeView (WidgetSet.insert wset, root, backgrnd, foregrnd, ch)
		     (item,
		      (nil, nil, nil, 0)))

		val box = Box.mkLayout root (Box.VtCenter outbox)

		fun process (allchannels, topchannels, count) =
		    (case CML.recv pvt of
			 Add items =>
			     (print "should add something\n";
			      process(allchannels, topchannels, count))
		       | Change items => 
			 let
			     fun change (i,v) =
				 (CML.send (List.nth (allchannels, i), IChange v)
				  handle Nth => raise BadIndex)
			 in
			     app change items;
			     process (allchannels, topchannels, count)
			 end
		       | ListAll out => 
			 (case topchannels of 
			      nil => CML.send(out, Report(nil,nil,out))
			    | head :: rest => CML.send(head, Report(nil, rest,out));
			  process (allchannels, topchannels, count)))
	    in
	      CML.spawn (fn () => process (allchannels, topchannels, count));
	      RL {
		  ch = pvt,
		  widget = Box.widgetOf box,
		  wset = wset,
		  evt = CML.recvEvt ch
		  }
	    end

    fun evtOf (RL{evt,...}) = evt
    fun widgetOf (RL{widget,...}) = widget

    fun setChosen (RL{wset,...}) arg = WidgetSet.setChosen wset arg
    fun setActive (RL{wset,...}) arg = WidgetSet.setActive wset arg
    fun setData (RL{ch,...}) arg = CML.send(ch, Change arg)

    fun getChosen (RL{wset,...}) = WidgetSet.getChosen wset
    fun getState (RL{wset,...}) = WidgetSet.getState wset

    fun addChild (RL{ch,...}) arg = CML.send(ch, Add arg)
    fun listItems (RL{ch,...})  = 
	let
	    val out = CML.channel ()
	in 
	    CML.send(ch, ListAll out);
	    case (CML.recv out) of
		Return [x] => x
	      | _ => (print "ack: recl list lost\n";
		      raise Match)
	end
    end
