
functor TreeEd(structure RL : REC_LIST) : TREE_ED =
    struct

	structure W = Widget
	structure S = Shell
	structure R = RL

	open CML Geometry W Shell Interact Frame MenuButton Shape Box SimpleMenu

	datatype 'a actions =
	    AL of {init : 'a,
		   edit : string * 'a * 'a option CML.chan * bool CML.chan -> S.shell,
		   done : 'a FieldStr.node option CML.chan,
		   die : bool CML.chan}

	datatype 'a treeinfo =
	    TI of {withMenu: bool,
		   actions: 'a actions,
		   elinit: 'a FieldStr.node}

	datatype 'a action =
	    Edit of (int * string * 'a)
	  | Add
	  | Finish of bool

	fun setup (root, deal, listw, 
		   AL{init, edit, done, die}) =
	    let
		fun zombie () =
		    (case recv deal of
			 x => zombie())

		fun process (Edit (code, key, data), (l, curfile)) =
		    let
			val editDone = channel()
			val editDie = channel()
			    
			val shell =  edit (key, data, editDone, editDie);

			fun loop () =
			    (case recv editDone of
				SOME result =>
				    (print "new value for ";
				     print (Int.toString code);
				     print "\n";
				     RL.setData listw [ (code, result) ])
			      | NONE =>
				    ();
			     RL.setActive listw [ (code, true) ];
			     RL.setChosen listw [ (code, false) ];
			     destroy shell;
			     exit ())
		    in
			spawn loop;
			(editDie :: l, curfile)
		    end
		  | process (Add, (l, curfile)) =
		    (print "add!\n";
		     (l, curfile))
		  | process (Finish b, (l, curfile)) =
		    (app (fn x => send (x, b)) l;
		     CML.send(done,
			      if b then 
				  SOME (RL.listItems listw)
			      else
				  NONE);
		     TraceCML.unwatch(getTid());
		     zombie())

		fun loop (lc) =
		    loop (process (select[recvEvt deal,
					  wrap (recvEvt die, Finish)],
				   lc))
	    in
		TraceCML.watch ("editor", spawn (fn () => loop (nil, "output")))
	    end

	fun mkMyRecList (root, deal, elt) =
	    let
		val recList =
		    RL.mkRList root {mode = RL.MultiSet,
				    backgrnd = NONE,
				    foregrnd = NONE,
				    item = elt}

		fun loop () =
		    (case sync (RL.evtOf recList) of
			 (RL.Set (code, key, element)) =>
			     (send(deal, Edit (code, key, element));
			      RL.setActive recList [ (code, false) ];
			      loop())
		       | (RL.Unset _) => 
			     (loop()))
	    in
		spawn loop;
		recList
	    end

	fun mkTreeMenu (root, deal) =
	    let
		val menu = 
		    MENU [MenuItem ("add", Add),
			  MenuItem ("ok", Finish true),
			  MenuItem ("cancel", Finish false)]

		val (widget, event) =
		    mkMenuButton root ("stuff", menu)

		fun loop () =
		    (case sync event of
			 x => (send(deal, x);
			       loop ()))
	    in
		spawn loop;
		widget
	    end

	fun mkTreeLayout (root, reclist, menu) =
	    let
		val glue = Glue{nat=5, min=1, max=NONE}
		val space = Glue{nat=5, min=5, max=SOME 5}
		val frame = 
		    mkFrame 
		    {color = SOME (EXB.blackOfScr (screenOf root)),
		     width = 2,
		     widget = RL.widgetOf reclist}
	    in
		mkLayout root
		(VtCenter [space,
			   HzCenter[WBox (mkRigid menu),
				    glue],
			   space,
			   HzCenter[space,
				    WBox (Frame.widgetOf frame),
				    space],
			   space])
	    end

	fun mkSimpleTreeD root (TI{withMenu,actions,elinit}) =
	    let
		val deal = channel ()
		val tlw = mkMyRecList (root, deal, elinit)
		val final = 
		    (case withMenu of
			 true =>
			     Box.widgetOf 
			     (mkTreeLayout(root, tlw, 		
					   mkTreeMenu (root, deal)))
		       | false =>
			     R.widgetOf tlw)
	    in
		setup(root,deal,tlw,actions);
		final
	    end

	fun mkTreeD root (ti) =
	    let
		val wrapper = mkSimpleTreeD root ti
		val shell = mkShell (wrapper, NONE, 
				     {win_name = SOME "leaving",
				      icon_name = SOME "leaving"})
	    in
		shell
	    end
    end

functor TreeNFieldEd(structure FE : FIELDS;
		     structure TE : TREE_ED) : TREE_N_FIELD_ED =
    struct
	structure W = Widget
	structure S = Shell
	structure FE = FE
	structure TE = TE

	datatype 'a actions =
	    AL of {init : 'a,
		   edit : string * 'a * 'a option CML.chan * bool CML.chan -> S.shell,
		   done : ('a FieldStr.node * string list) option CML.chan,
		   die : bool CML.chan}

	datatype 'a tfinfo =
	    TFI of {actions: 'a actions,
		    name: string,
		    minwidth : int option,
		    fields : FE.field list,
		    tree : 'a FieldStr.node}

	datatype 'a which =
	    TDone of 'a FieldStr.node option
	  | FDone of string list option
	  | Die of bool

	fun setup (root, 
		   treedone: 'a FieldStr.node option CML.chan, treedie,
		   fielddone: string list option CML.chan, fielddie,
		   mydone, mydie) =
	    let
		fun zombie x = 
		    (case CML.select [CML.wrap(CML.recvEvt treedone, fn _ => ()),
				      CML.wrap(CML.recvEvt fielddone, fn _ => ()),
				      CML.wrap(CML.recvEvt mydie, fn _ => ())] of
			 _ => zombie x)

		fun killTree (f, b) =
		    (CML.send (treedie, b);
		     if b then
			 f (CML.recv treedone)
		     else
			 f (NONE))

		fun killField (f, b) =
		    (CML.send (fielddie, b);
		     if b then
			 f (CML.recv fielddone)
		     else
			 f (NONE))
		
		fun sendout false _ _ =
		    (zombie ())
		  | sendout _ NONE _ =
		    (CML.send(mydone, NONE);
		     zombie ())
		  | sendout _ _  NONE =
		    (CML.send(mydone, NONE);
		     zombie ())
		  | sendout _ (SOME t) (SOME f) =
		    (CML.send(mydone, SOME(t,f));
		     zombie ())

		fun process (TDone x, replyp) = 
		    killField (sendout replyp x, 
			       case x of NONE => false | _ => true)

		  | process (FDone y, replyp) = 
		    killTree (fn x => sendout replyp x y,
			       case y of NONE => false | _ => true)

		  | process (Die b, _) = 
		    killField (fn x => process(FDone x,b), b)
				
		fun loop () = 
		    process(CML.select [CML.wrap(CML.recvEvt treedone, TDone),
					CML.wrap(CML.recvEvt fielddone, FDone),
					CML.wrap(CML.recvEvt mydie, Die)],
			    true)
			

	    in
		CML.spawn loop
	    end

	fun mkFieldPart (root, TFI{name, minwidth, fields, ...}) =
	    let
		val fieldDone = CML.channel ()
		val fieldDie = CML.channel ()
		val fi = FE.FI{name=name,
			       minwidth=minwidth,
			       done=fieldDone,
			       die=fieldDie,
			       fields=fields,
			       action = NONE}
	    in
		(FE.mkSimpleFields (root, fi),
		 fieldDone,
		 fieldDie)
	    end

	fun mkFieldViewPart (root, TFI{name, minwidth, fields, ...}) =
	    let
		val fieldDone = CML.channel ()
		val fieldDie = CML.channel ()
		val fi = FE.FI{name=name,
			       minwidth=minwidth,
			       done=fieldDone,
			       die=fieldDie,
			       fields=fields,
			       action = NONE}
	    in
		(FE.mkSimpleFieldView (root, fi),
		 fieldDone,
		 fieldDie)
	    end


	fun mkTreePart (root, TFI{actions=AL{init,edit,...},tree,...}) =
	    let
		val treeDone = CML.channel ()
		val treeDie = CML.channel ()
		val ti = TE.TI{withMenu = false,
			       actions = TE.AL{init = init,
					       edit = edit,
					       done = treeDone,
					       die = treeDie},
			       elinit = tree}
	    in
		(TE.mkSimpleTreeD root ti,
		 treeDone,
		 treeDie)
	    end

	fun mkTFLayout (root, fieldwidget, treewidget) =
	    let
		val space = Box.Glue{nat=5, min=5, max=SOME 5}

		val treelabel = 
		    Label.mkLabel root {label = "tree:",
					font = NONE,
					foregrnd = NONE,
					backgrnd = NONE,
					align = W.HLeft}

		val black  = 
		    SOME (W.EXB.colorOfScr
			  (W.screenOf root)
			  (W.EXB.CMS_Name "black"))

		val frame =
		    Frame.mkFrame {color = black,
				   width = 2,
				   widget = treewidget}
					
	    in
		Box.widgetOf (Box.mkLayout root
			      (Box.HzCenter[space,
					    Box.WBox fieldwidget,
					    space,
					    Box.VtLeft[space,
						       Box.WBox
						       (Label.widgetOf treelabel),
						       space,
						       Box.WBox (Frame.widgetOf frame),
						       space]]))
	    end

	fun mkTreeNField root (tfinfo as TFI{actions=AL{done,die,...},name,...}) =
	    let
		val (tw, tdone, tdie) = mkTreePart(root, tfinfo)
		val (fw, fdone, fdie) = mkFieldPart(root, tfinfo)
		val final = mkTFLayout (root, fw, tw)
		val shell = Shell.mkShell(final, NONE,
					  {win_name = SOME name,
					   icon_name = SOME name})
	    in
		setup(root, tdone, tdie, fdone, fdie, done, die);
		shell
	    end

	fun mkTreeNFieldView root (tfinfo as TFI{actions=AL{done,die,...},name,...}) =
	    let
		val (tw, tdone, tdie) = mkTreePart(root, tfinfo)
		val (fw, fdone, fdie) = mkFieldViewPart(root, tfinfo)
		val final = mkTFLayout (root, fw, tw)
		val shell = Shell.mkShell(final, NONE,
					  {win_name = SOME name,
					   icon_name = SOME name})

		fun wait () =
		    case (CML.select [CML.wrap(CML.recvEvt tdone, fn _ => [fdie]),
				      CML.wrap(CML.recvEvt fdone, fn _ => [tdie]),
				      CML.wrap(CML.recvEvt die, fn _ => [fdie,tdie])]) of
			l => (app (fn x => CML.send(x, false)) l;
			      CML.send(done, NONE); 
			      wait())
	    in
		CML.spawn wait;
		shell
	    end
		    
    end

